from typing import List, Union

import numpy as np
from tqdm import tqdm

from cv_pipeliner.batch_generators.image_data import BatchGeneratorImageData
from cv_pipeliner.core.data import ImageData
from cv_pipeliner.core.inferencer import Inferencer
from cv_pipeliner.inference_models.embedder.core import EmbedderModel


class EmbedderInferencer(Inferencer):
    def __init__(self, model: EmbedderModel):
        assert isinstance(model, EmbedderModel)
        super().__init__(model)

    def predict(
        self,
        images_data_gen: Union[ImageData, BatchGeneratorImageData, "DataLoader"],
        batch_size_default: int = 16,
        disable_tqdm: bool = False,
    ) -> List[np.ndarray]:
        try:
            from torch.utils.data import DataLoader
        except ImportError:
            DataLoader = None
        use_dataloader_torch = DataLoader is not None and isinstance(images_data_gen, DataLoader)

        if use_dataloader_torch:
            pass
        else:
            if isinstance(images_data_gen, list):
                images_data_gen = BatchGeneratorImageData(images_data_gen, batch_size=batch_size_default)
            assert isinstance(images_data_gen, BatchGeneratorImageData)

        pred_embeddings = []
        for images in tqdm(images_data_gen, total=len(images_data_gen), disable=disable_tqdm):
            if use_dataloader_torch:
                images = images[0]
            else:
                images = [image_data.image for image_data in images]

            embedding = self.model.predict(input=images)
            pred_embeddings.extend(embedding)
        return pred_embeddings



import logging
import os
import sys
from datetime import datetime
from pathlib import Path

LOGS_DIRECTORY = Path(__file__).parent / "__logs__"

logger = logging.getLogger("cv-pipeliner")

if os.environ.get("CV_PIPELINER_LOGGING", False):
    LOGS_DIRECTORY.mkdir(exist_ok=True, parents=True)
    LOG_FILENAME = datetime.now().strftime("%Y-%m-%d_%Hh.logs")

    formatter = logging.Formatter("%(asctime)s [%(name)s] [%(levelname)s] %(message)s")

    file_handler = logging.FileHandler(LOGS_DIRECTORY / LOG_FILENAME, encoding="utf-8")
    file_handler.setFormatter(formatter)
    file_handler.setLevel(logging.INFO)
    logger.addHandler(file_handler)

    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.setFormatter(formatter)
    stream_handler.setLevel(logging.INFO)
    logger.addHandler(stream_handler)
    logger.propagate = False

    try:
        from tensorflow import get_logger as tf_get_logger

        tf_logger = tf_get_logger()
        tf_logger.addHandler(file_handler)
        tf_logger.propagate = False
    except ModuleNotFoundError:
        pass

    for handler in tf_logger.handlers:
        handler.setFormatter(formatter)



import abc
import copy
from typing import Any, List

import numpy as np


class BatchGenerator(abc.ABC):
    def __init__(self, data: List[Any], batch_size: int, use_not_caught_elements_as_last_batch: bool):
        """
        Base generator class for data.

        Argument use_not_caught_elements_as_last_batch is used
        when len(data) is not divisible by batch_size, but we need to get
        all data by using the generator (e.g. for inference).
        For training use_not_caught_elements_as_last_batch should be set as False.
        """
        self.data = np.array(data, dtype=object)
        self.batch_size = batch_size
        self.indexes = np.arange(len(self.data))
        if use_not_caught_elements_as_last_batch:
            self._len = int(np.ceil(len(self.data) / self.batch_size))
        else:
            assert len(self.data) // self.batch_size != 0
            self._len = len(self.data) // self.batch_size

    def __len__(self) -> int:
        return self._len

    def __getitem__(self, index) -> List:
        indexes = self.indexes[index * self.batch_size : (index + 1) * self.batch_size]
        batch = copy.deepcopy(self.data[indexes])
        return batch

    def __iter__(self):
        for item in (self[i] for i in range(len(self))):
            yield item
            del item



import io
from pathlib import Path
from typing import Dict, List, Union

import fsspec

from cv_pipeliner.core.data import BboxData, ImageData
from cv_pipeliner.core.data_converter import DataConverter
from cv_pipeliner.utils.images_datas import combine_mask_polygons_to_one_polygon
from cv_pipeliner.utils.imagesize import get_image_size


class YOLODataConverter(DataConverter):
    def __init__(self, class_names: List[str]):
        super().__init__()
        assert len(set(class_names)) == len(class_names), "There are duplicates in 'class_names'. Remove them."
        self.class_names = class_names
        self.class_name_to_idx = {class_name: idx for idx, class_name in enumerate(self.class_names)}
        self.idx_to_class_name = {idx: class_name for idx, class_name in enumerate(self.class_names)}

    def get_annot_from_image_data(self, image_data: ImageData) -> List[str]:
        image_data = self.filter_image_data(image_data)
        width, height = image_data.get_image_size()
        txt_results = []
        for bbox_data in image_data.bboxes_data:
            w = bbox_data.xmax - bbox_data.xmin
            h = bbox_data.ymax - bbox_data.ymin
            xcenter = bbox_data.xmin + w / 2
            ycenter = bbox_data.ymin + h / 2
            xcenter, w = round(xcenter / width, 6), round(w / width, 6)
            ycenter, h = round(ycenter / height, 6), round(h / height, 6)
            idx = self.class_name_to_idx[bbox_data.label]
            txt_results.append(f"{idx} {xcenter} {ycenter} {w} {h}")
        return txt_results

    @DataConverter.assert_image_data
    def get_image_data_from_annot(
        self,
        image_path: Union[str, Path],
        annot: Union[Path, str, Dict, fsspec.core.OpenFile, List[str]],
    ) -> ImageData:
        if isinstance(annot, str) or isinstance(annot, Path):
            with fsspec.open(annot, "r", encoding="utf8") as f:
                annots = f.read()
        elif isinstance(annot, fsspec.core.OpenFile):
            with annot as f:
                annots = f.read()
        elif isinstance(annot, io.IOBase):
            annots = annot.read()
            if isinstance(annots, bytes):
                annots = annots.decode()
        elif isinstance(annot, List):
            annots = "\n".join(annot)

        width, height = get_image_size(image_path)
        bboxes_data = []
        for line in annots.strip().split("\n"):
            if line == "":
                continue
            idx, xcenter, ycenter, w, h = line.split(" ")
            label = self.idx_to_class_name[int(idx)]
            xcenter, ycenter, w, h = float(xcenter), float(ycenter), float(w), float(h)
            xcenter, w = xcenter * width, w * width
            ycenter, h = ycenter * height, h * height
            bboxes_data.append(
                BboxData(
                    xmin=xcenter - w / 2,
                    ymin=ycenter - h / 2,
                    xmax=xcenter + w / 2,
                    ymax=ycenter + h / 2,
                    label=label,
                )
            )

        return ImageData(image_path=image_path, bboxes_data=bboxes_data)


class YOLOMasksDataConverter(DataConverter):
    """
    Converter ImageData to YOLO Keypoints and YOLO Keypoints to ImageData
    """

    def __init__(self, class_names: List[str]):
        super().__init__()
        assert len(set(class_names)) == len(class_names), "There are duplicates in 'class_names'. Remove them."
        self.class_names = class_names
        self.class_name_to_idx = {class_name: idx for idx, class_name in enumerate(self.class_names)}
        self.idx_to_class_name = {idx: class_name for idx, class_name in enumerate(self.class_names)}

    def get_annot_from_image_data(self, image_data: ImageData) -> List[str]:
        image_data = self.filter_image_data(image_data)
        width, height = image_data.get_image_size()
        # txt_coors_results = []
        txt_keypoints_results = []
        for idx, bbox_data in enumerate(image_data.bboxes_data):
            idx = self.class_name_to_idx[bbox_data.label]
            mask = combine_mask_polygons_to_one_polygon(bbox_data.mask)
            box_keypoins = f"{idx}"
            for x, y in mask:
                box_keypoins += f" {round(x/width, 5)} {round(y/height, 5)}"
            txt_keypoints_results.append(box_keypoins)

        return txt_keypoints_results  # txt_coors_results,

    @DataConverter.assert_image_data
    def get_image_data_from_annot(
        self,
        image_path: Union[str, Path],
        annot: Union[Path, str, Dict, fsspec.core.OpenFile, List[str]],
    ) -> ImageData:
        if isinstance(annot, str) or isinstance(annot, Path):
            with fsspec.open(annot, "r", encoding="utf8") as f:
                annots = f.read()
        elif isinstance(annot, fsspec.core.OpenFile):
            with annot as f:
                annots = f.read()
        elif isinstance(annot, io.IOBase):
            annots = annot.read()
            if isinstance(annots, bytes):
                annots = annots.decode()
        elif isinstance(annot, List):
            annots = "\n".join(annot)

        width, height = get_image_size(image_path)
        bboxes_data = []
        for line in annots.strip().split("\n"):
            if line == "":
                continue

            elements = line.split()
            idx = elements[0]
            label = self.idx_to_class_name[int(idx)]
            points = elements[1:]
            scaled_points = [
                (float(points[i]) * width if i % 2 == 0 else float(points[i]) * height) for i in range(len(points))
            ]
            # Определяем минимальные и максимальные координаты для бокса
            xs, ys = scaled_points[0::2], scaled_points[1::2]  # x и y координаты точек

            xmin, xmax, ymin, ymax = min(xs), max(xs), min(ys), max(ys)

            mask = [[scaled_points[i], scaled_points[i + 1]] for i in range(0, len(scaled_points), 2)]

            bboxes_data.append(
                BboxData(
                    xmin=xmin,
                    ymin=ymin,
                    xmax=xmax,
                    ymax=ymax,
                    mask=[mask],
                    label=label,
                )
            )

        return ImageData(image_path=image_path, bboxes_data=bboxes_data)



import cv2

from cv_pipeliner.tracking.bbox_utils import coco_bboxes_to_voc, voc_bboxes_to_coco


class OpenCVTracker:
    def __init__(self, initial_bboxes, initial_frame):
        self.multiTracker = cv2.MultiTracker_create()
        formated_bboxes = voc_bboxes_to_coco(initial_bboxes.copy())
        for bbox in formated_bboxes:
            self.multiTracker.add(cv2.TrackerMedianFlow_create(), initial_frame, bbox)

    def update(self, frame_pixels):
        success, tracked_bboxes = self.multiTracker.update(frame_pixels)
        tracked_bboxes = coco_bboxes_to_voc(tracked_bboxes, frame_pixels.shape)
        return tracked_bboxes



from importlib.metadata import version

from packaging.version import Version

if Version(version("pydantic")) < Version("2.0.0"):
    from pydantic import BaseModel as PydanticBaseModel
else:
    from pydantic.v1 import BaseModel as PydanticBaseModel

import abc
import importlib
import sys
import tempfile
from pathlib import Path
from typing import Callable, List, Optional, Tuple, Type, Union

import fsspec
import numpy as np


class BaseModel(PydanticBaseModel):
    class Config:
        arbitrary_types_allowed = True


class ModelSpec(BaseModel):
    id: str = None

    @abc.abstractproperty
    def inference_model_cls(self) -> Type["InferenceModel"]:
        pass

    def load(self, **kwargs) -> Union["InferenceModel", "Inferencer"]:
        inference_model = self.inference_model_cls.get_loaded_model_by_id(self.id)
        if inference_model is None:
            inference_model = self.inference_model_cls(model_spec=self, **kwargs)
        return inference_model


class InferenceModel(abc.ABC):
    """
    Low-level class for models.
    To define the model, we need to create an object, then load checkpoint from given model_spec.

    Example:
        model_spec = ModelSpec(...)
        inference_model = InferenceModel(model_spec)
        output = inference_model.predict(input)

    2nd way:
        model_spec = ModelSpec(...)
        inference_model = model_spec.load()
        output = inference_model.predict(input)


    "input" and "output" types should be defined in the inheritance of this class.

    """

    _loaded_models: List[ModelSpec] = []

    @staticmethod
    def get_loaded_model_by_id(id: Optional[str]) -> Optional["InferenceModel"]:
        if id is None:
            return None
        ids = [model._model_spec.id for model in InferenceModel._loaded_models]
        if id in ids:
            return InferenceModel._loaded_models[ids.index(id)]
        return None

    def __init__(self, model_spec: ModelSpec):
        self._model_spec = model_spec
        if self._model_spec.id is not None:
            InferenceModel._loaded_models.append(self)

    def __del__(self):
        ids = [model._model_spec.id for model in InferenceModel._loaded_models]
        if self._model_spec.id is not None and self._model_spec.id in ids:
            InferenceModel._loaded_models.pop(ids.index(self._model_spec.id))

    @abc.abstractmethod
    def predict(self, input):
        pass

    @abc.abstractmethod
    def preprocess_input(self, input):
        pass

    @abc.abstractproperty
    def input_size(self) -> Tuple[int, int]:
        pass

    @property
    def model_spec(self):
        return self._model_spec


def get_preprocess_input_from_script_file(script_file: Union[str, Path]) -> Callable[[List[np.ndarray]], np.ndarray]:
    with fsspec.open(script_file, "r") as src:
        script_code = src.read()
    with tempfile.TemporaryDirectory() as tmpdirname:
        tmpdirname = Path(tmpdirname)
        module_folder = tmpdirname / "module"
        module_folder.mkdir()
        script_file = module_folder / f"preprocess_input_{tmpdirname.name}.py"
        with open(script_file, "w") as out:
            out.write(script_code)
        sys.path.append(str(script_file.parent.absolute()))
        module = importlib.import_module(script_file.stem)
        importlib.reload(module)
        sys.path.pop()
    return module.preprocess_input



import contextlib
import os
from typing import Dict, List

import numpy as np
import pandas as pd

from cv_pipeliner.core.data import ImageData
from cv_pipeliner.metrics.image_data_matching import ImageDataMatching


def count_coco_metrics(
    true_images_data: List[ImageData], raw_pred_images_data: List[ImageData] = None, class_names: List[str] = None
) -> Dict:
    from object_detection.core.standard_fields import (
        DetectionResultFields,
        InputDataFields,
    )
    from object_detection.metrics import coco_evaluation

    if class_names is None:
        cocoevaluator = coco_evaluation.CocoDetectionEvaluator(categories=[{"id": 1, "name": "Label"}])
    else:
        cocoevaluator = coco_evaluation.CocoDetectionEvaluator(
            categories=[{"id": i + 1, "name": class_name} for i, class_name in enumerate(class_names)]
        )

    for i, (true_image_data, raw_pred_image_data) in enumerate(zip(true_images_data, raw_pred_images_data)):
        true_bboxes_data = true_image_data.bboxes_data
        raw_pred_bboxes_data = raw_pred_image_data.bboxes_data

        true_bboxes = [
            (true_bbox_data.ymin, true_bbox_data.xmin, true_bbox_data.ymax, true_bbox_data.xmax)
            for true_bbox_data in true_bboxes_data
        ]
        raw_pred_bboxes = [
            (raw_pred_bbox_data.ymin, raw_pred_bbox_data.xmin, raw_pred_bbox_data.ymax, raw_pred_bbox_data.xmax)
            for raw_pred_bbox_data in raw_pred_bboxes_data
        ]
        raw_pred_scores = [raw_pred_bbox_data.detection_score for raw_pred_bbox_data in raw_pred_bboxes_data]

        groundtruth_boxes = np.array(true_bboxes, dtype=np.float32).reshape(-1, 4)
        if class_names is None:
            groundtruth_classes = np.array([1] * len(true_bboxes))
        else:
            groundtruth_classes = np.array([class_names.index(bbox_data.label) + 1 for bbox_data in true_bboxes_data])
        groundtruth_dict = {
            InputDataFields.groundtruth_boxes: groundtruth_boxes,
            InputDataFields.groundtruth_classes: groundtruth_classes,
        }
        cocoevaluator.add_single_ground_truth_image_info(image_id=i, groundtruth_dict=groundtruth_dict)

        detection_boxes = np.array(raw_pred_bboxes, dtype=np.float32).reshape(-1, 4)
        detection_scores = np.array(raw_pred_scores, dtype=np.float32)
        if class_names is None:
            detection_classes = np.array([1] * len(detection_boxes))
        else:
            detection_classes = np.array([class_names.index(bbox_data.label) + 1 for bbox_data in raw_pred_bboxes_data])
        detections_dict = {
            DetectionResultFields.detection_boxes: detection_boxes,
            DetectionResultFields.detection_scores: detection_scores,
            DetectionResultFields.detection_classes: detection_classes,
        }
        cocoevaluator.add_single_detected_image_info(image_id=i, detections_dict=detections_dict)

    with open(os.devnull, "w") as f, contextlib.redirect_stdout(f):
        coco_metrics = cocoevaluator.evaluate()
    return coco_metrics


df_detection_metrics_columns = ["value"]


def get_df_detection_metrics(
    true_images_data: List[ImageData],
    pred_images_data: List[ImageData],
    minimum_iou: float,
    raw_pred_images_data: List[ImageData] = None,
) -> pd.DataFrame:
    """
    Returns detection metrics (precision, recall, f1_score, mAP).
    """

    assert len(true_images_data) == len(pred_images_data)

    images_data_matchings = [
        ImageDataMatching(true_image_data, pred_image_data, minimum_iou)
        for true_image_data, pred_image_data in zip(true_images_data, pred_images_data)
    ]
    TP = np.sum([image_data_matching.get_detection_TP() for image_data_matching in images_data_matchings])
    FP = np.sum([image_data_matching.get_detection_FP() for image_data_matching in images_data_matchings])
    FN = np.sum([image_data_matching.get_detection_FN() for image_data_matching in images_data_matchings])
    iou_mean = np.mean(
        [
            bbox_data_matching.iou
            for image_data_matching in images_data_matchings
            for bbox_data_matching in image_data_matching.bboxes_data_matchings
            if bbox_data_matching.iou is not None
        ]
    )
    accuracy = TP / max(TP + FN + FN, 1e-6)
    precision = TP / max(TP + FP, 1e-6)
    recall = TP / max(TP + FN, 1e-6)
    f1_score = 2 * precision * recall / max(precision + recall, 1e-6)
    coco_metrics = {}
    if raw_pred_images_data is not None:
        try:
            coco_metrics = count_coco_metrics(true_images_data, raw_pred_images_data)
        except ModuleNotFoundError:
            pass

    df_detection_metrics = pd.DataFrame(
        {
            "images_support": len(true_images_data),
            "support": [TP + FN],
            "TP": [TP],
            "FP": [FP],
            "FN": [FN],
            "iou_mean": [iou_mean],
            "accuracy": [accuracy],
            "precision": [precision],
            "recall": [recall],
            "f1_score": [f1_score],
            **coco_metrics,
        },
        dtype=object,
    ).T
    df_detection_metrics.columns = df_detection_metrics_columns

    return df_detection_metrics



import abc
from typing import List, Tuple, Type

import numpy as np

from cv_pipeliner.core.inference_model import InferenceModel, ModelSpec

Label = str
Score = float

Labels_Top_N = List[List[Label]]
Scores_Top_N = List[List[Score]]

ClassificationInput = List[np.ndarray]
ClassificationOutput = Tuple[Labels_Top_N, Scores_Top_N]


class ClassificationModelSpec(ModelSpec):
    @abc.abstractproperty
    def inference_model_cls(self) -> Type["ClassificationModel"]:
        pass

    def load_classification_inferencer(self) -> "ClassificationInferencer":
        from cv_pipeliner.inferencers.classification import ClassificationInferencer

        return ClassificationInferencer(self.load())


class ClassificationModel(InferenceModel):
    def __init__(self, model_spec: ClassificationModelSpec):
        assert isinstance(model_spec, ClassificationModelSpec)
        super().__init__(model_spec)

    @abc.abstractmethod
    def predict(self, input: ClassificationInput, top_n: int = 1) -> ClassificationOutput:
        pass

    @abc.abstractmethod
    def preprocess_input(self, input):
        pass

    @abc.abstractproperty
    def input_size(self) -> Tuple[int, int]:
        pass

    @abc.abstractproperty
    def class_names(self) -> List[str]:
        pass



import json
from pathlib import Path
from typing import Dict, Union

import fsspec

from cv_pipeliner.core.data import BboxData, ImageData
from cv_pipeliner.core.data_converter import DataConverter


class SuperviselyDataConverter(DataConverter):
    @DataConverter.assert_image_data
    def get_image_data_from_annot(
        self, image_path: Union[str, Path, fsspec.core.OpenFile], annot: Union[Path, str, Dict, fsspec.core.OpenFile]
    ) -> ImageData:
        if isinstance(annot, str) or isinstance(annot, Path):
            with fsspec.open(annot, "r", encoding="utf8") as f:
                annot = json.load(f)
        if isinstance(annot, fsspec.core.OpenFile):
            with annot as f:
                annot = json.load(f)

        bboxes_data = []
        for obj in annot["objects"]:
            (xmin, ymin), (xmax, ymax) = obj["points"]["exterior"]
            label = obj["tags"][0]["name"] if obj["tags"] else None
            bboxes_data.append(BboxData(image_path=image_path, xmin=xmin, ymin=ymin, xmax=xmax, ymax=ymax, label=label))

        image_data = ImageData(image_path=image_path, bboxes_data=bboxes_data)

        return image_data

    def get_annot_from_image_data(self, image_data: ImageData) -> Dict:
        image_data = self.filter_image_data(image_data)
        image = image_data.open_image()
        height, width, _ = image.shape
        annot = {
            "description": "",
            "tags": [],
            "size": {"height": height, "width": width},
            "objects": [
                {
                    "description": "",
                    "geometryType": "rectangle",
                    "tags": [
                        {
                            "name": str(bbox_data.label),
                            "value": None,
                        }
                    ],
                    "classTitle": "bbox",
                    "points": {
                        "exterior": [
                            [int(bbox_data.xmin), int(bbox_data.ymin)],
                            [int(bbox_data.xmax), int(bbox_data.ymax)],
                        ],
                        "interior": [],
                    },
                }
                for bbox_data in image_data.bboxes_data
            ],
        }
        return annot



import abc
import json
from pathlib import Path
from typing import Dict, List, Union

import fsspec
from joblib import Parallel, delayed
from tqdm import tqdm

from cv_pipeliner.core.data import BboxData, ImageData


class DataConverter(abc.ABC):
    def __init__(self):
        pass

    def filter_image_data(self, image_data: ImageData) -> ImageData:
        if image_data is None:
            return None
        looked_bboxes = set()
        new_bboxes_data = []
        for bbox_data in image_data.bboxes_data:
            xmin, ymin, xmax, ymax = bbox_data.xmin, bbox_data.ymin, bbox_data.xmax, bbox_data.ymax
            xmin, ymin, xmax, ymax = int(xmin), int(ymin), int(xmax), int(ymax)

            if (xmin, ymin, xmax, ymax) in looked_bboxes:
                # logger.warning(
                #     f"Repeated bbox detected at image {bbox_data.image_path}: "
                #     f"(xmin, ymin, xmax, ymax) = {(xmin, xmin, ymax, xmax)}. Skipping."
                # )
                continue
            else:
                looked_bboxes.add((xmin, ymin, xmax, ymax))

            if xmin >= xmax or ymin >= ymax or xmin < 0 or ymin < 0:
                # logger.warning(
                #     f"Wrong annotation at image {bbox_data.image_path}: "
                #     f"incorrect bbox (xmin, ymin, xmax, ymax): {(xmin, ymin, xmax, ymax)} "
                #     "(xmin >= xmax or ymin >= ymax or xmin < 0 or ymin < 0). Skipping."
                # )
                continue

            new_bboxes_data.append(bbox_data)

        image_data = ImageData(
            image_path=image_data.image_path,
            image=image_data.image,
            bboxes_data=new_bboxes_data,
            additional_info=image_data.additional_info,
            keypoints=image_data.keypoints,
            label=image_data.label,
        )

        return image_data

    def assert_image_data(fn):
        def wrapped(data_converter: "DataConverter", *args, **kwargs) -> ImageData:
            image_data = fn(data_converter, *args, **kwargs)
            image_data = data_converter.filter_image_data(image_data)
            return image_data

        return wrapped

    def assert_images_data(fn):
        def wrapped(data_converter: "DataConverter", *args, **kwargs) -> List[ImageData]:
            images_data = fn(data_converter, *args, **kwargs)
            images_data = [data_converter.filter_image_data(image_data) for image_data in images_data]
            return images_data

        return wrapped

    @abc.abstractmethod
    @assert_image_data
    def get_image_data_from_annot(
        self, image_path: Union[str, Path], annot: Union[Path, str, Dict, fsspec.core.OpenFile]
    ) -> ImageData:
        pass

    def get_images_data_from_annots(
        self,
        image_paths: List[Union[str, Path]],
        annots: Union[List[Union[Path, str, Dict]], Union[Path, str, Dict]],
        n_jobs: int = 8,
        disable_tqdm: bool = False,
    ) -> List[ImageData]:
        if isinstance(annots, str) or isinstance(annots, Path):
            with fsspec.open(annots, "r", encoding="utf8") as f:
                annots = json.load(f)
        if isinstance(annots, List):
            assert len(image_paths) == len(annots)
        images_data = Parallel(n_jobs=n_jobs, prefer="threads")(
            delayed(self.get_image_data_from_annot)(image_path=image_path, annot=annot)
            for image_path, annot in tqdm(zip(image_paths, annots), total=len(image_paths), disable=disable_tqdm)
        )
        images_data = [image_data for image_data in images_data if image_data is not None]
        return images_data

    def get_n_bboxes_data_from_annots(
        self, image_paths: List[Union[str, Path]], annots: Union[List[Union[Path, str, Dict]], Union[Path, str, Dict]]
    ) -> List[List[BboxData]]:
        images_data = self.get_images_data_from_annots(image_paths=image_paths, annots=annots)
        n_bboxes_data = [image_data.bboxes_data for image_data in images_data]
        return n_bboxes_data

    def get_annot_from_image_data(self, image_data: ImageData) -> Dict:
        return {}

    def get_annot_from_images_data(self, images_data: ImageData) -> Dict:
        annots = [self.get_annot_from_image_data(image_data) for image_data in images_data]
        return annots

    def get_annot_from_n_bboxes_data(
        self,
        image_paths: List[Union[str, Path]],
        n_bboxes_data: List[List[BboxData]],
    ) -> List[List[BboxData]]:
        images_data = [
            ImageData(image_path=image_path, bboxes_data=bboxes_data)
            for image_path, bboxes_data in zip(image_paths, n_bboxes_data)
        ]
        return self.get_annot_from_images_data(images_data)



from typing import Dict, List

import numpy as np
import pandas as pd

from cv_pipeliner.core.data import ImageData
from cv_pipeliner.metrics.image_data_matching import BboxDataMatching, ImageDataMatching


def _count_errors_types_and_get_pipeline_metrics_per_class(
    images_data_matchings: List[ImageDataMatching],
    labels: List[str],
    extra_bbox_label: str,
    filter_by_true_labels: bool,
) -> Dict:
    pipeline_metrics_per_class = {}
    if filter_by_true_labels:
        images_data_matchings = [
            ImageDataMatching(
                true_image_data=image_data_matching.true_image_data,
                pred_image_data=image_data_matching.pred_image_data,
                minimum_iou=image_data_matching.minimum_iou,
                extra_bbox_label=image_data_matching.extra_bbox_label,
                bboxes_data_matchings=[
                    BboxDataMatching(
                        true_bbox_data=bbox_data_matching.true_bbox_data,
                        pred_bbox_data=bbox_data_matching.pred_bbox_data,
                        extra_bbox_label=bbox_data_matching.extra_bbox_label,
                    )
                    for bbox_data_matching in image_data_matching.bboxes_data_matchings
                    if (
                        bbox_data_matching.true_bbox_data is not None
                        and bbox_data_matching.true_bbox_data.label in labels
                    )
                ],
            )
            for image_data_matching in images_data_matchings
        ]
    true_labels = np.array(
        [
            bbox_data_matching.true_bbox_data.label
            for image_data_matching in images_data_matchings
            for bbox_data_matching in image_data_matching.bboxes_data_matchings
            if bbox_data_matching.true_bbox_data is not None
        ]
    )
    for class_name in labels:
        images_support_by_class_name = np.sum(
            [
                any(
                    [
                        bbox_data_matching.true_bbox_data.label == class_name
                        for bbox_data_matching in image_data_matching.bboxes_data_matchings
                        if bbox_data_matching.true_bbox_data is not None
                    ]
                )
                for image_data_matching in images_data_matchings
            ]
        )
        support_by_class_name = np.sum(true_labels == class_name)
        TP_by_class_name = np.sum(
            [image_data_matching.get_pipeline_TP(label=class_name) for image_data_matching in images_data_matchings]
        )
        FP_by_class_name = np.sum(
            [image_data_matching.get_pipeline_FP(label=class_name) for image_data_matching in images_data_matchings]
        )
        if class_name != extra_bbox_label:
            TP_extra_bbox_by_class_name = np.sum(
                [
                    image_data_matching.get_pipeline_TP_extra_bbox(label=class_name)
                    for image_data_matching in images_data_matchings
                ]
            )
            FP_extra_bbox_by_class_name = np.sum(
                [
                    image_data_matching.get_pipeline_FP_extra_bbox(label=class_name)
                    for image_data_matching in images_data_matchings
                ]
            )
            FN_extra_bbox_by_class_name = np.sum(
                [
                    image_data_matching.get_pipeline_FN_extra_bbox(label=class_name)
                    for image_data_matching in images_data_matchings
                ]
            )
        else:
            TP_extra_bbox_by_class_name = None
            FP_extra_bbox_by_class_name = None
            FN_extra_bbox_by_class_name = None
        FN_by_class_name = np.sum(
            [image_data_matching.get_pipeline_FN(label=class_name) for image_data_matching in images_data_matchings]
        )
        TP_extra_bbox_in_precision_numerator = 0 if TP_extra_bbox_by_class_name is None else TP_extra_bbox_by_class_name
        FP_extra_bbox_in_precision_denominator = (
            0 if FP_extra_bbox_by_class_name is None else FP_extra_bbox_by_class_name
        )
        FN_extra_bbox_in_precision_denominator = (
            0 if FN_extra_bbox_by_class_name is None else FN_extra_bbox_by_class_name
        )
        precision_by_class_name = (TP_by_class_name + TP_extra_bbox_in_precision_numerator) / max(
            TP_by_class_name + FP_by_class_name + FP_extra_bbox_in_precision_denominator, 1e-6
        )
        recall_by_class_name = (TP_by_class_name + TP_extra_bbox_in_precision_numerator) / max(
            TP_by_class_name + FN_by_class_name + FN_extra_bbox_in_precision_denominator, 1e-6
        )
        f1_score_by_class_name = (
            2
            * precision_by_class_name
            * recall_by_class_name
            / (max(precision_by_class_name + recall_by_class_name, 1e-6))
        )
        ious = [
            bbox_data_matching.iou
            for image_data_matching in images_data_matchings
            for bbox_data_matching in image_data_matching.bboxes_data_matchings
            if bbox_data_matching.iou is not None
            and bbox_data_matching.true_bbox_data is not None
            and bbox_data_matching.true_bbox_data.label == class_name
        ]
        iou_mean = np.mean(ious) if len(ious) > 0 else 0
        pipeline_metrics_per_class[class_name] = {
            "images_support": images_support_by_class_name,
            "support": support_by_class_name,
            "TP": TP_by_class_name,
            "FP": FP_by_class_name,
            "FN": FN_by_class_name,
            "TP (extra bbox)": TP_extra_bbox_by_class_name,
            "FP (extra bbox)": FP_extra_bbox_by_class_name,
            "FN (extra bbox)": FN_extra_bbox_by_class_name,
            "iou_mean": iou_mean,
            "precision": precision_by_class_name,
            "recall": recall_by_class_name,
            "f1_score": f1_score_by_class_name,
        }
    return pipeline_metrics_per_class


def _add_metrics_to_dict(
    pipeline_metrics_per_class: Dict,
    pipeline_metrics: Dict,
    labels: List[str],
    images_support: int,
    prefix_caption: str = "",
    postfix_caption: str = "",
):
    supports = [pipeline_metrics_per_class[class_name]["support"] for class_name in labels]
    support = np.sum(supports)
    TP = np.sum([pipeline_metrics_per_class[class_name]["TP"] for class_name in labels])
    FP = np.sum([pipeline_metrics_per_class[class_name]["FP"] for class_name in labels])
    FN = np.sum([pipeline_metrics_per_class[class_name]["FN"] for class_name in labels])
    TP_extra_bbox = np.sum(
        [
            pipeline_metrics_per_class[class_name]["TP (extra bbox)"]
            for class_name in labels
            if pipeline_metrics_per_class[class_name]["TP (extra bbox)"] is not None
        ]
    )
    FP_extra_bbox = np.sum(
        [
            pipeline_metrics_per_class[class_name]["FP (extra bbox)"]
            for class_name in labels
            if pipeline_metrics_per_class[class_name]["FP (extra bbox)"] is not None
        ]
    )
    FN_extra_bbox = np.sum(
        [
            pipeline_metrics_per_class[class_name]["FN (extra bbox)"]
            for class_name in labels
            if pipeline_metrics_per_class[class_name]["FN (extra bbox)"] is not None
        ]
    )
    iou_mean = np.mean(
        [
            pipeline_metrics_per_class[class_name]["iou_mean"]
            for class_name in labels
            if pipeline_metrics_per_class[class_name]["iou_mean"] is not None
        ]
    )
    accuracy = (TP + TP_extra_bbox) / max(TP + FP + FN + TP_extra_bbox + FP_extra_bbox, 1e-6)
    micro_average_precision = (TP + TP_extra_bbox) / max(TP + FP + FP_extra_bbox, 1e-6)
    micro_average_recall = TP / max(TP + FN, 1e-6)
    micro_average_f1_score = (
        2 * micro_average_precision * micro_average_recall / (max(micro_average_precision + micro_average_recall, 1e-6))
    )
    precisions = [pipeline_metrics_per_class[class_name]["precision"] for class_name in labels]
    recalls = [pipeline_metrics_per_class[class_name]["recall"] for class_name in labels]
    f1_scores = [pipeline_metrics_per_class[class_name]["f1_score"] for class_name in labels]
    macro_average_precision = np.average(precisions) if len(precisions) > 0 else np.nan
    weighted_average_precision = np.average(precisions, weights=supports) if len(precisions) > 0 else np.nan
    macro_average_recall = np.average(recalls) if len(recalls) > 0 else np.nan
    weighted_average_recall = np.average(recalls, weights=supports) if len(recalls) > 0 else np.nan
    macro_average_f1_score = np.average(f1_scores) if len(f1_scores) > 0 else np.nan
    weighted_average_f1_score = np.average(f1_scores, weights=supports) if len(f1_scores) > 0 else np.nan
    sum_support = np.sum(supports)
    pipeline_metrics[f"{prefix_caption}accuracy{postfix_caption}"] = {
        "images_support": images_support,
        "support": support,
        "TP": TP,
        "FP": FP,
        "FN": FN,
        "TP (extra bbox)": TP_extra_bbox,
        "FP (extra bbox)": FP_extra_bbox,
        "FN (extra bbox)": FN_extra_bbox,
        "value": accuracy,
    }
    pipeline_metrics[f"{prefix_caption}iou_mean{postfix_caption}"] = {"support": support, "value": iou_mean}
    pipeline_metrics[f"{prefix_caption}micro_average{postfix_caption}"] = {
        "images_support": images_support,
        "support": sum_support,
        "TP": TP,
        "FP": FP,
        "FN": FN,
        "TP (extra bbox)": TP_extra_bbox,
        "FP (extra bbox)": FP_extra_bbox,
        "FN (extra bbox)": FN_extra_bbox,
        "precision": micro_average_precision,
        "recall": micro_average_recall,
        "f1_score": micro_average_f1_score,
    }
    pipeline_metrics[f"{prefix_caption}macro_average{postfix_caption}"] = {
        "images_support": images_support,
        "support": sum_support,
        "TP": TP,
        "FP": FP,
        "FN": FN,
        "TP (extra bbox)": TP_extra_bbox,
        "FP (extra bbox)": FP_extra_bbox,
        "FN (extra bbox)": FN_extra_bbox,
        "precision": macro_average_precision,
        "recall": macro_average_recall,
        "f1_score": macro_average_f1_score,
    }
    pipeline_metrics[f"{prefix_caption}weighted_average{postfix_caption}"] = {
        "images_support": images_support,
        "support": sum_support,
        "TP": TP,
        "FP": FP,
        "FN": FN,
        "TP (extra bbox)": TP_extra_bbox,
        "FP (extra bbox)": FP_extra_bbox,
        "FN (extra bbox)": FN_extra_bbox,
        "precision": weighted_average_precision,
        "recall": weighted_average_recall,
        "f1_score": weighted_average_f1_score,
    }


df_pipeline_metrics_columns = [
    "images_support",
    "support",
    "value",
    "TP",
    "FP",
    "FN",
    "TP (extra bbox)",
    "FP (extra bbox)",
    "FN (extra bbox)",
    "precision",
    "recall",
    "f1_score",
]


def get_df_pipeline_metrics(
    true_images_data: List[ImageData],
    pred_images_data: List[ImageData],
    minimum_iou: float,
    extra_bbox_label: str = "trash (extra bbox)",
    pseudo_class_names: List[str] = [],
    known_class_names: List[str] = None,
) -> pd.DataFrame:
    """
    Returns pipdline metrics (accuracy, precision, recall, f1_score), including metrics per class..
    """
    images_support = len(true_images_data)
    images_data_matchings = [
        ImageDataMatching(
            true_image_data=true_image_data,
            pred_image_data=pred_image_data,
            minimum_iou=minimum_iou,
            extra_bbox_label=extra_bbox_label,
        )
        for true_image_data, pred_image_data in list(zip(true_images_data, pred_images_data))
    ]
    true_labels = np.array([bbox_data.label for image_data in true_images_data for bbox_data in image_data.bboxes_data])
    pred_labels = np.array([bbox_data.label for image_data in pred_images_data for bbox_data in image_data.bboxes_data])
    all_class_names = np.unique(np.concatenate([true_labels, pred_labels]))
    pipeline_metrics_per_class_all_class_names = _count_errors_types_and_get_pipeline_metrics_per_class(
        images_data_matchings=images_data_matchings,
        labels=all_class_names,
        extra_bbox_label=extra_bbox_label,
        filter_by_true_labels=False,
    )
    pipeline_metrics = {}
    for class_name in all_class_names:
        pipeline_metrics[class_name] = pipeline_metrics_per_class_all_class_names[class_name]
    TP_extra_bbox = np.sum(
        [
            image_data_matching.get_pipeline_TP_extra_bbox(label=extra_bbox_label)
            for image_data_matching in images_data_matchings
        ]
    )
    FP_extra_bbox = np.sum(
        [
            image_data_matching.get_pipeline_FP_extra_bbox(label=extra_bbox_label)
            for image_data_matching in images_data_matchings
        ]
    )
    FN_extra_bbox = np.sum(
        [
            image_data_matching.get_pipeline_FN_extra_bbox(label=extra_bbox_label)
            for image_data_matching in images_data_matchings
        ]
    )
    precision_extra_bbox = TP_extra_bbox / max(TP_extra_bbox + FP_extra_bbox, 1e-6)
    recall_extra_bbox = TP_extra_bbox / max(TP_extra_bbox + FN_extra_bbox, 1e-6)
    f1_score_extra_bbox = (
        2 * precision_extra_bbox * recall_extra_bbox / (max(precision_extra_bbox + recall_extra_bbox, 1e-6))
    )
    extra_bbox_label_caption = f"{extra_bbox_label} (extra bbox)" if extra_bbox_label is not None else "extra bbox"
    pipeline_metrics[extra_bbox_label_caption] = {
        "images_support": images_support,
        "support": TP_extra_bbox + FP_extra_bbox + FN_extra_bbox,
        "TP": 0,
        "FP": 0,
        "FN": 0,
        "TP (extra bbox)": TP_extra_bbox,
        "FP (extra bbox)": FP_extra_bbox,
        "FN (extra bbox)": FN_extra_bbox,
        "precision": precision_extra_bbox,
        "recall": recall_extra_bbox,
        "f1_score": f1_score_extra_bbox,
    }

    class_names_without_pseudo_classes = list(set(all_class_names) - set(pseudo_class_names))
    _add_metrics_to_dict(
        pipeline_metrics_per_class=pipeline_metrics_per_class_all_class_names,
        pipeline_metrics=pipeline_metrics,
        labels=all_class_names,
        images_support=images_support,
        prefix_caption="all_",
    )
    _add_metrics_to_dict(
        pipeline_metrics_per_class=pipeline_metrics_per_class_all_class_names,
        pipeline_metrics=pipeline_metrics,
        labels=class_names_without_pseudo_classes,
        images_support=images_support,
        prefix_caption="all_",
        postfix_caption="_without_pseudo_classes",
    )
    if known_class_names is not None:
        known_class_names = list(set(all_class_names).intersection(set(known_class_names)))
        known_class_names_without_pseudo_classes = list(set(known_class_names) - set(pseudo_class_names))
        _add_metrics_to_dict(
            pipeline_metrics_per_class=pipeline_metrics_per_class_all_class_names,
            pipeline_metrics=pipeline_metrics,
            labels=known_class_names,
            images_support=images_support,
            prefix_caption="known_",
        )
        _add_metrics_to_dict(
            pipeline_metrics_per_class=pipeline_metrics_per_class_all_class_names,
            pipeline_metrics=pipeline_metrics,
            labels=known_class_names_without_pseudo_classes,
            images_support=images_support,
            prefix_caption="known_",
            postfix_caption="_without_pseudo_classes",
        )

    df_pipeline_metrics = pd.DataFrame(pipeline_metrics, dtype=object).T
    df_pipeline_metrics = df_pipeline_metrics[df_pipeline_metrics_columns]
    df_pipeline_metrics.sort_values(by="support", ascending=False, inplace=True)
    if known_class_names is not None:
        df_pipeline_metrics.loc[all_class_names, "known"] = df_pipeline_metrics.loc[all_class_names].index.isin(
            known_class_names
        )
        df_pipeline_metrics.loc[all_class_names, "pseudo"] = df_pipeline_metrics.loc[all_class_names].index.isin(
            pseudo_class_names
        )

    return df_pipeline_metrics



import tempfile
from typing import Tuple, Type, Union

import fsspec
import numpy as np
from pathy import Pathy

from cv_pipeliner.inference_models.keypoints_regressor.core import (
    KeypointsRegressorInput,
    KeypointsRegressorModel,
    KeypointsRegressorModelSpec,
    KeypointsRegressorOutput,
)
from cv_pipeliner.inference_models.keypoints_regressor.utils.mmpose_utils import (
    bbox_xywh2cs,
    keypoints_from_heatmaps,
    preprocess,
    top_down_affine,
)


class MMPose_KeypointsRegressorModelSpec_TFLite(KeypointsRegressorModelSpec):
    model_path: Union[str, Pathy]  # can be also tf.keras.Model

    @property
    def inference_model_cls(self) -> Type["MMPose_KeypointsRegressorModel"]:
        from cv_pipeliner.inference_models.keypoints_regressor.mmpose import (
            MMPose_KeypointsRegressorModel,
        )

        return MMPose_KeypointsRegressorModel


INPUT_TYPE_TO_DTYPE = {
    "image_tensor": np.uint8,
    "float_image_tensor": np.float32,
    "encoded_image_string_tensor": np.uint8,
}


class MMPose_KeypointsRegressorModel(KeypointsRegressorModel):
    def _load_tensorflow_KeypointsRegressor_model_spec(self, model_spec: MMPose_KeypointsRegressorModelSpec_TFLite):
        import tensorflow as tf

        model_openfile = fsspec.open(model_spec.model_path, "rb")
        temp_file = tempfile.NamedTemporaryFile()
        with model_openfile as src:
            temp_file.write(src.read())
        model_path = Pathy.fluid(temp_file.name)
        temp_files_cleanup = temp_file.close

        self.model = tf.lite.Interpreter(str(model_path))
        self.model.allocate_tensors()
        self.input_details = self.model.get_input_details()[0]
        self.input_index = self.input_details["index"]
        self.input_dtype = self.input_details["dtype"]
        self.output_index = self.model.get_output_details()[0]["index"]

        temp_files_cleanup()

    def __init__(self, model_spec: MMPose_KeypointsRegressorModelSpec_TFLite):
        super().__init__(model_spec)

        if isinstance(model_spec, MMPose_KeypointsRegressorModelSpec_TFLite):
            self._load_tensorflow_KeypointsRegressor_model_spec(model_spec)
            self._raw_predict = self._raw_predict_tensorflow
        else:
            raise ValueError(
                f"MMPose_KeypointsRegressorModel got unknown MMPose_KeypointsRegressorModelSpec: {type(model_spec)}"
            )

    def _raw_predict_tensorflow(self, image: np.ndarray):
        import tensorflow as tf

        image = tf.convert_to_tensor(image, dtype=self.input_dtype)
        self.model.set_tensor(self.input_index, image)
        self.model.invoke()
        heatmaps = self.model.get_tensor(self.output_index)
        return heatmaps

    def predict(self, input: KeypointsRegressorInput) -> KeypointsRegressorOutput:
        n_keypoints = []
        for image in input:
            height, width, _ = image.shape
            center, scale = bbox_xywh2cs([0, 0, width, height], aspect_ratio=192 / 256, padding=1.25, pixel_std=200)
            image = top_down_affine(image, 0, center, scale, [192, 256], False)
            image = preprocess(image)
            heatmaps = self._raw_predict(image)
            keypoints = keypoints_from_heatmaps(
                heatmaps,
                [center],
                [scale],
                unbiased=False,
                post_process="default",
                kernel=11,
                valid_radius_factor=0.0546875,
                use_udp=False,
                target_type="GaussianHeatmap",
            )[0][0]
            n_keypoints.append(keypoints)

        n_keypoints = np.array(n_keypoints)
        return n_keypoints

    def preprocess_input(self, input: KeypointsRegressorInput):
        return input

    @property
    def input_size(self) -> Tuple[int, int]:
        return self.model_spec.input_size



import abc
from typing import List, Type

import numpy as np

from cv_pipeliner.core.inference_model import InferenceModel, ModelSpec

EmbedderInput = List[np.ndarray]
EmbedderOutput = List[np.ndarray]


class EmbedderModelSpec(ModelSpec):
    @abc.abstractproperty
    def inference_model_cls(self) -> Type["EmbedderModel"]:
        pass

    def load_embedder_inferencer(self) -> "EmbedderInferencer":
        from cv_pipeliner.inferencers.embedder import EmbedderInferencer

        return EmbedderInferencer(self.load())


class EmbedderModel(InferenceModel):
    def __init__(self, model_spec: EmbedderModelSpec):
        assert isinstance(model_spec, EmbedderModelSpec)
        super().__init__(model_spec)

    @abc.abstractmethod
    def predict(
        self,
        input: EmbedderInput,
    ) -> EmbedderOutput:
        pass

    @abc.abstractmethod
    def preprocess_input(self, input):
        pass

    @abc.abstractproperty
    def input_size(self) -> int:
        pass



import tempfile
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import Callable, List, Literal, Tuple, Type, Union

import fsspec
import numpy as np
import requests
from pathy import Pathy

from cv_pipeliner.core.inference_model import get_preprocess_input_from_script_file
from cv_pipeliner.inference_models.embedder.core import (
    EmbedderInput,
    EmbedderModel,
    EmbedderModelSpec,
    EmbedderOutput,
)
from cv_pipeliner.utils.files import copy_files_from_directory_to_temp_directory
from cv_pipeliner.utils.images import get_image_b64


class TensorFlow_EmbedderModelSpec(EmbedderModelSpec):
    input_size: Union[Tuple[int, int], List[int]]
    model_path: Union[str, Pathy]  # can be also tf.keras.Model
    saved_model_type: Literal["tf.saved_model", "tf.keras", "tf.keras.Model", "tflite", "tflite_one_image_per_batch"]
    preprocess_input: Union[Callable[[List[np.ndarray]], np.ndarray], str, Path, None] = None

    @property
    def inference_model_cls(self) -> Type["Tensorflow_EmbedderModel"]:
        from cv_pipeliner.inference_models.embedder.tensorflow import (
            Tensorflow_EmbedderModel,
        )

        return Tensorflow_EmbedderModel


class TensorFlow_EmbedderModelSpec_TFServing(EmbedderModelSpec):
    url: str
    input_type: Literal["image_tensor", "float_image_tensor", "encoded_image_string_tensor"]
    input_name: str
    input_size: Union[Tuple[int, int], List[int]]
    preprocess_input: Union[Callable[[List[np.ndarray]], np.ndarray], str, Path, None] = None

    @property
    def inference_model_cls(self) -> Type["Tensorflow_EmbedderModel"]:
        from cv_pipeliner.inference_models.embedder.tensorflow import (
            Tensorflow_EmbedderModel,
        )

        return Tensorflow_EmbedderModel


INPUT_TYPE_TO_DTYPE = {
    "image_tensor": np.uint8,
    "float_image_tensor": np.float32,
    "encoded_image_string_tensor": np.uint8,
}


class Tensorflow_EmbedderModel(EmbedderModel):
    def _load_tensorflow_embedder_model_spec(self, model_spec: TensorFlow_EmbedderModelSpec):
        import tensorflow as tf

        if model_spec.saved_model_type in ["tf.keras", "tf.saved_model", "tflite", "tflite_one_image_per_batch"]:
            model_openfile = fsspec.open(model_spec.model_path, "rb")
            if model_openfile.fs.isdir(model_openfile.path):
                temp_folder = copy_files_from_directory_to_temp_directory(directory=model_spec.model_path)
                model_path = Pathy.fluid(temp_folder.name)
                temp_files_cleanup = temp_folder.cleanup
            else:
                temp_file = tempfile.NamedTemporaryFile()
                with model_openfile as src:
                    temp_file.write(src.read())
                model_path = Pathy.fluid(temp_file.name)
                temp_files_cleanup = temp_file.close

            if model_spec.saved_model_type in "tf.keras":
                self.model = tf.keras.models.load_model(str(model_path), compile=False)
                self.input_dtype = np.float32
            elif model_spec.saved_model_type == "tf.saved_model":
                self.loaded_model = tf.saved_model.load(str(model_path))  # to protect from gc
                self.model = self.loaded_model.signatures["serving_default"]
                self.input_dtype = np.float32
            elif model_spec.saved_model_type in ["tflite", "tflite_one_image_per_batch"]:
                self.model = tf.lite.Interpreter(str(model_path))
                input_details = self.model.get_input_details()[0]
                self.input_index = input_details["index"]
                self.input_dtype = input_details["dtype"]
                self.output_index = self.model.get_output_details()[0]["index"]

            temp_files_cleanup()

        elif model_spec.saved_model_type == "tf.keras.Model":
            self.model = model_spec.model_path
            self.input_dtype = np.float32
        else:
            raise ValueError(
                "Tensorflow_EmbedderModel got unknown saved_model_type "
                f"in TensorFlow_EmbedderModelSpec: {model_spec.saved_model_type}"
            )

    def __init__(self, model_spec: Union[TensorFlow_EmbedderModelSpec, TensorFlow_EmbedderModelSpec_TFServing]):
        super().__init__(model_spec)

        if isinstance(model_spec, TensorFlow_EmbedderModelSpec):
            self._load_tensorflow_embedder_model_spec(model_spec)
            self._raw_predict = self._raw_predict_tensorflow
        elif isinstance(model_spec, TensorFlow_EmbedderModelSpec_TFServing):
            self.input_dtype = INPUT_TYPE_TO_DTYPE[model_spec.input_type]
            # Wake up the service
            try:
                self._raw_predict_kfserving(images=np.zeros((1, *self.input_size, 3)), timeout=1.0)
            except requests.exceptions.ReadTimeout:
                pass
            self._raw_predict = self._raw_predict_kfserving
        else:
            raise ValueError(f"Tensorflow_EmbedderModel got unknown EmbedderModelSpec: {type(model_spec)}")

        if isinstance(model_spec.preprocess_input, str) or isinstance(model_spec.preprocess_input, Path):
            self._preprocess_input = get_preprocess_input_from_script_file(script_file=model_spec.preprocess_input)
        else:
            if model_spec.preprocess_input is None:
                self._preprocess_input = lambda x: x
            else:
                self._preprocess_input = model_spec.preprocess_input

    def _raw_predict_tensorflow(self, images: np.ndarray):
        import tensorflow as tf

        if self.model_spec.saved_model_type == "tf.saved_model":
            input_tensor = tf.convert_to_tensor(images, dtype=self.input_dtype)
            raw_predictions_batch = self.model(input_tensor)
            if isinstance(raw_predictions_batch, dict):
                key = list(raw_predictions_batch)[0]
                raw_predictions_batch = np.array(raw_predictions_batch[key])
        elif self.model_spec.saved_model_type in ["tf.keras", "tf.keras.Model"]:
            if len(images) > 0:
                raw_predictions_batch = self.model.predict(images)
            else:
                raw_predictions_batch = []
        elif self.model_spec.saved_model_type == "tflite":
            images = tf.convert_to_tensor(np.array(images), dtype=self.input_dtype)
            self.model.resize_tensor_input(0, [len(images), *self.input_size, 3])
            self.model.allocate_tensors()
            self.model.set_tensor(self.input_index, images)
            self.model.invoke()
            raw_predictions_batch = self.model.get_tensor(self.output_index)
        elif self.model_spec.saved_model_type == "tflite_one_image_per_batch":
            raw_predictions_batch = []
            for image in images:
                height, width, _ = image.shape
                image_tensor = tf.convert_to_tensor(np.array([image]), dtype=self.input_dtype)
                self.model.resize_tensor_input(0, [1, height, width, 3])
                self.model.allocate_tensors()
                self.model.set_tensor(self.input_index, image_tensor)
                self.model.invoke()
                raw_predictions_batch.append(self.model.get_tensor(self.output_index)[0])

        raw_predictions_batch = np.array(raw_predictions_batch)
        return raw_predictions_batch

    def _raw_predict_kfserving(self, images: np.ndarray, timeout: Union[float, None] = None):
        if self.model_spec.input_type in ["float_image_tensor", "image_tensor"]:
            input_data = {
                "inputs": {
                    self.model_spec.input_name: [np.array(image).astype(self.input_dtype).tolist() for image in images]
                }
            }
        elif self.model_spec.input_type == "encoded_image_string_tensor":
            input_data = {
                "instances": [
                    {self.model_spec.input_name: {"b64": get_image_b64(image, "JPEG", quality=95)}} for image in images
                ]
            }
        response = requests.post(url=self.model_spec.url, json=input_data, timeout=timeout)
        try:
            output_dict = response.json()
        except JSONDecodeError:
            raise ValueError(f"Failed to decode JSON. Response content: {response.content}")
        if not response.ok:
            raise ValueError(f"Response is not ok: {response.status_code=}; {response.content=}")
        if "outputs" in output_dict:
            raw_predictions_batch = np.array(output_dict["outputs"])
        elif "predictions" in output_dict:
            raw_predictions_batch = np.array(output_dict["predictions"])

        return raw_predictions_batch

    def predict(self, input: EmbedderInput) -> EmbedderOutput:
        input = self.preprocess_input(input)
        predictions = self._raw_predict(input)
        return predictions

    def preprocess_input(self, input: EmbedderInput):
        return self._preprocess_input(input)

    @property
    def input_size(self) -> Tuple[int, int]:
        return self.model_spec.input_size



import json
import tempfile
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, Type, Union

import fsspec
import numpy as np
import requests
from pathy import Pathy

from cv_pipeliner.core.inference_model import get_preprocess_input_from_script_file
from cv_pipeliner.inference_models.classification.core import (
    ClassificationInput,
    ClassificationModel,
    ClassificationModelSpec,
    ClassificationOutput,
)
from cv_pipeliner.utils.files import copy_files_from_directory_to_temp_directory
from cv_pipeliner.utils.images import get_image_b64


class TensorFlow_ClassificationModelSpec(ClassificationModelSpec):
    input_size: Optional[Union[Tuple[Optional[int], Optional[int]], List[Optional[int]]]]
    class_names: Union[List[str], str, Path]
    model_path: Union[str, Pathy]  # can be also tf.keras.Model
    saved_model_type: Literal["tf.saved_model", "tf.keras", "tf.keras.Model", "tflite", "tflite_one_image_per_batch"]
    preprocess_input: Union[Callable[[List[np.ndarray]], np.ndarray], str, Path, None] = None
    device: Optional[str] = None  # https://www.tensorflow.org/guide/gpu#manual_device_placement
    custom_objects: Optional[Dict[str, Any]] = None

    @property
    def inference_model_cls(self) -> Type["Tensorflow_ClassificationModel"]:
        from cv_pipeliner.inference_models.classification.tensorflow import (
            Tensorflow_ClassificationModel,
        )

        return Tensorflow_ClassificationModel


class TensorFlow_ClassificationModelSpec_TFServing(ClassificationModelSpec):
    url: str
    input_type: Literal["image_tensor", "float_image_tensor", "encoded_image_string_tensor"]
    input_name: str
    input_size: Optional[Union[Tuple[Optional[int], Optional[int]], List[Optional[int]]]]
    class_names: Union[List[str], str, Path]
    preprocess_input: Union[Callable[[List[np.ndarray]], np.ndarray], str, Path, None] = None

    @property
    def inference_model_cls(self) -> Type["Tensorflow_ClassificationModel"]:
        from cv_pipeliner.inference_models.classification.tensorflow import (
            Tensorflow_ClassificationModel,
        )

        return Tensorflow_ClassificationModel


INPUT_TYPE_TO_DTYPE = {
    "image_tensor": np.uint8,
    "float_image_tensor": np.float32,
    "encoded_image_string_tensor": np.uint8,
}


class Tensorflow_ClassificationModel(ClassificationModel):
    def _load_tensorflow_classification_model_spec(self, model_spec: TensorFlow_ClassificationModelSpec):
        import tensorflow as tf

        if model_spec.saved_model_type in ["tf.keras", "tf.saved_model", "tflite", "tflite_one_image_per_batch"]:
            model_openfile = fsspec.open(model_spec.model_path, "rb")
            if model_openfile.fs.isdir(model_openfile.path):
                temp_folder = copy_files_from_directory_to_temp_directory(directory=model_spec.model_path)
                model_path = Pathy.fluid(temp_folder.name)
                temp_files_cleanup = temp_folder.cleanup
            else:
                temp_file = tempfile.NamedTemporaryFile()
                with model_openfile as src:
                    temp_file.write(src.read())
                model_path = Pathy.fluid(temp_file.name)
                temp_files_cleanup = temp_file.close

            if model_spec.device is not None:
                self.tf_device = tf.device(model_spec.device)
                self.tf_device.__enter__()
            else:
                self.tf_device = None
            try:
                if model_spec.saved_model_type in "tf.keras":
                    self.model = tf.keras.models.load_model(
                        str(model_path), compile=False, custom_objects=model_spec.custom_objects
                    )
                    self.input_dtype = np.float32
                elif model_spec.saved_model_type == "tf.saved_model":
                    self.loaded_model = tf.saved_model.load(str(model_path))  # to protect from gc
                    self.model = self.loaded_model.signatures["serving_default"]
                    self.input_dtype = np.float32
                elif model_spec.saved_model_type in ["tflite", "tflite_one_image_per_batch"]:
                    self.model = tf.lite.Interpreter(str(model_path))
                    input_details = self.model.get_input_details()[0]
                    self.input_index = input_details["index"]
                    self.input_dtype = input_details["dtype"]
                    self.output_index = self.model.get_output_details()[0]["index"]
            except Exception as e:
                raise e
            finally:
                if self.tf_device is not None:
                    self.tf_device.__exit__()

            temp_files_cleanup()

        elif model_spec.saved_model_type == "tf.keras.Model":
            self.model = model_spec.model_path
            self.input_dtype = np.float32
        else:
            raise ValueError(
                "Tensorflow_ClassificationModel got unknown saved_model_type "
                f"in TensorFlow_ClassificationModelSpec: {model_spec.saved_model_type}"
            )

    def __init__(
        self, model_spec: Union[TensorFlow_ClassificationModelSpec, TensorFlow_ClassificationModelSpec_TFServing]
    ):
        super().__init__(model_spec)

        if isinstance(model_spec.class_names, str) or isinstance(model_spec.class_names, Path):
            with fsspec.open(model_spec.class_names, "r", encoding="utf-8") as out:
                self._class_names = json.load(out)
        else:
            self._class_names = model_spec.class_names

        if isinstance(model_spec, TensorFlow_ClassificationModelSpec):
            self._load_tensorflow_classification_model_spec(model_spec)
            self._raw_predict = self._raw_predict_tensorflow
        elif isinstance(model_spec, TensorFlow_ClassificationModelSpec_TFServing):
            self.input_dtype = INPUT_TYPE_TO_DTYPE[model_spec.input_type]
            # Wake up the service
            try:
                self._raw_predict_kfserving(images=np.zeros((1, *self.input_size, 3)), timeout=1.0)
            except requests.exceptions.ReadTimeout:
                pass
            self._raw_predict = self._raw_predict_kfserving
        else:
            raise ValueError(f"Tensorflow_ClassificationModel got unknown ClassificationModelSpec: {type(model_spec)}")

        if isinstance(model_spec.preprocess_input, str) or isinstance(model_spec.preprocess_input, Path):
            self._preprocess_input = get_preprocess_input_from_script_file(script_file=model_spec.preprocess_input)
        else:
            if model_spec.preprocess_input is None:
                self._preprocess_input = lambda x: x
            else:
                self._preprocess_input = model_spec.preprocess_input

        self.id_to_class_name = np.array([class_name for class_name in self._class_names])

    def _raw_predict_tensorflow(self, images: np.ndarray):
        import tensorflow as tf

        if self.tf_device is not None:
            self.tf_device.__enter__()
        try:
            if self.model_spec.saved_model_type == "tf.saved_model":
                input_tensor = tf.convert_to_tensor(images, dtype=self.input_dtype)
                raw_predictions_batch = self.model(input_tensor)
                if isinstance(raw_predictions_batch, dict):
                    key = list(raw_predictions_batch)[0]
                    raw_predictions_batch = np.array(raw_predictions_batch[key])
            elif self.model_spec.saved_model_type in ["tf.keras", "tf.keras.Model"]:
                if len(images) > 0:
                    raw_predictions_batch = self.model.predict(images, verbose=0)
                else:
                    raw_predictions_batch = []
            elif self.model_spec.saved_model_type == "tflite":
                images = tf.convert_to_tensor(np.array(images), dtype=self.input_dtype)
                self.model.resize_tensor_input(0, [len(images), *self.input_size, 3])
                self.model.allocate_tensors()
                self.model.set_tensor(self.input_index, images)
                self.model.invoke()
                raw_predictions_batch = self.model.get_tensor(self.output_index)
            elif self.model_spec.saved_model_type == "tflite_one_image_per_batch":
                raw_predictions_batch = []
                for image in images:
                    height, width, _ = image.shape
                    image_tensor = tf.convert_to_tensor(np.array([image]), dtype=self.input_dtype)
                    self.model.resize_tensor_input(0, [1, height, width, 3])
                    self.model.allocate_tensors()
                    self.model.set_tensor(self.input_index, image_tensor)
                    self.model.invoke()
                    raw_predictions_batch.append(self.model.get_tensor(self.output_index)[0])
            raw_predictions_batch = np.array(raw_predictions_batch).reshape(-1, len(self.class_names))
        except Exception as e:
            raise e
        finally:
            if self.tf_device is not None:
                self.tf_device.__exit__()

        return raw_predictions_batch

    def _raw_predict_kfserving(self, images: np.ndarray, timeout: Union[float, None] = None):
        if self.model_spec.input_type in ["float_image_tensor", "image_tensor"]:
            input_data = {
                "inputs": {
                    self.model_spec.input_name: [np.array(image).astype(self.input_dtype).tolist() for image in images]
                }
            }
        elif self.model_spec.input_type == "encoded_image_string_tensor":
            input_data = {
                "instances": [
                    {self.model_spec.input_name: {"b64": get_image_b64(image, "JPEG", quality=95)}} for image in images
                ]
            }
        response = requests.post(url=self.model_spec.url, json=input_data, timeout=timeout)
        try:
            output_dict = response.json()
        except JSONDecodeError:
            raise ValueError(f"Failed to decode JSON. Response content: {response.content}")
        if not response.ok:
            raise ValueError(f"Response is not ok: {response.status_code=}; {response.content=}")
        if "outputs" in output_dict:
            raw_predictions_batch = np.array(output_dict["outputs"])
        elif "predictions" in output_dict:
            raw_predictions_batch = np.array(output_dict["predictions"])

        return raw_predictions_batch

    def predict(self, input: ClassificationInput, top_n: int = 1) -> ClassificationOutput:
        input = self.preprocess_input(input)
        predictions = self._raw_predict(input)
        max_scores_top_n_idxs = (-np.array(predictions)).argsort(axis=1)[:, :top_n]
        id_to_class_names_repeated = np.repeat(a=self.id_to_class_name[None, ...], repeats=len(input), axis=0)
        pred_labels_top_n = np.take_along_axis(id_to_class_names_repeated, max_scores_top_n_idxs, axis=1)
        pred_scores_top_n = np.take_along_axis(predictions, max_scores_top_n_idxs, axis=1)

        return pred_labels_top_n, pred_scores_top_n

    def preprocess_input(self, input: ClassificationInput):
        return self._preprocess_input(input)

    @property
    def input_size(self) -> Tuple[int, int]:
        return self.model_spec.input_size

    @property
    def class_names(self) -> List[str]:
        return self._class_names



from typing import Any, List

import ipywidgets as widgets
import matplotlib.pyplot as plt
from IPython.display import Image, clear_output, display


class JupyterVisualizer:
    def __init__(
        self,
        images,
        images_names,
        display_fn=lambda x: display(Image(x)),
        choices: List[Any] = [],
        choices_description: str = "Choice:",
        choices2: List[Any] = [],
        choices2_description: str = "Choice2:",
    ):
        self.current_image = 0
        self.display_fn = display_fn
        self.out = widgets.Output()
        self.images_dropdown = widgets.Dropdown(options=list(zip(images_names, images)), description="Image:")
        hbox_children = [self.images_dropdown]
        if choices:
            self.choices = widgets.Dropdown(options=choices, description=choices_description)
            self.choices.change_options = False
            hbox_children.append(self.choices)
        else:
            self.choices = None
        if choices2:
            self.choices2 = widgets.Dropdown(options=choices2, description=choices2_description)
            self.choices2.change_options = False
            hbox_children.append(self.choices2)
        else:
            self.choices2 = None
        self.hbox = widgets.HBox(children=tuple(hbox_children))
        self.vbox = widgets.VBox(children=(self.hbox, self.out))

    def visualize(self):
        display(self.vbox)

        @self.out.capture()
        def update_out(change):
            with self.out:
                clear_output(wait=True)
                self.display_fn(self.images_dropdown.value)
                plt.show()

        @self.out.capture()
        def on_change(change):
            if not self.choices.change_options:
                update_out(self)

        @self.out.capture()
        def on_change2(change):
            if not self.choices2.change_options:
                update_out(self)

        self.images_dropdown.observe(update_out, names=["value"])
        if self.choices:
            self.choices.observe(on_change, names=["value"])
        if self.choices2:
            self.choices2.observe(on_change2, names=["value"])
        update_out(self)



import json
from pathlib import Path
from typing import Dict, Type, Union

import fsspec

from cv_pipeliner.core.data import ImageData
from cv_pipeliner.core.data_converter import DataConverter


class JSONDataConverter(DataConverter):
    def __init__(self, image_data_cls: Type[ImageData] = ImageData):
        self.image_data_cls = image_data_cls

    def get_annot_from_image_data(self, image_data: ImageData) -> Dict:
        image_data = self.filter_image_data(image_data)
        return json.loads(image_data.json())

    @DataConverter.assert_image_data
    def get_image_data_from_annot(
        self, image_path: Union[str, Path], annot: Union[Path, str, Dict, fsspec.core.OpenFile]
    ) -> ImageData:
        return self.image_data_cls.from_json(annot, image_path=image_path)



from typing import Any, Callable, Dict, List, Tuple, Union

from tqdm import tqdm

from cv_pipeliner.batch_generators.image_data import BatchGeneratorImageData
from cv_pipeliner.core.data import BboxData, ImageData
from cv_pipeliner.core.inferencer import Inferencer
from cv_pipeliner.inference_models.pipeline import PipelineModel


class PipelineInferencer(Inferencer):
    def __init__(self, model: PipelineModel):
        assert isinstance(model, PipelineModel)
        super().__init__(model)

    def _postprocess_predictions(
        self,
        images_data: List[ImageData],
        n_pred_bboxes: List[List[Tuple[int, int, int, int]]],
        n_k_pred_keypoints: List[List[Tuple[int, int]]],
        n_k_pred_masks: List[List[List[List[Tuple[int, int]]]]],
        n_pred_detection_scores: List[List[float]],
        n_pred_labels_top_n: List[List[List[str]]],
        n_pred_classification_scores_top_n: List[List[List[float]]],
        open_images_in_images_data: bool,
        open_cropped_images_in_bboxes_data: bool,
    ) -> List[ImageData]:
        pred_images_data = []
        for (
            image_data,
            pred_bboxes,
            k_pred_keypoints,
            k_pred_masks,
            pred_detection_scores,
            pred_labels_top_n,
            pred_classification_scores_top_n,
        ) in zip(
            images_data,
            n_pred_bboxes,
            n_k_pred_keypoints,
            n_k_pred_masks,
            n_pred_detection_scores,
            n_pred_labels_top_n,
            n_pred_classification_scores_top_n,
        ):
            bboxes_data = []
            for (
                pred_bbox,
                pred_keypoints,
                pred_mask,
                pred_detection_score,
                pred_label_top_n,
                pred_classification_score_top_n,
            ) in zip(
                pred_bboxes,
                k_pred_keypoints,
                k_pred_masks,
                pred_detection_scores,
                pred_labels_top_n,
                pred_classification_scores_top_n,
            ):
                xmin, ymin, xmax, ymax = pred_bbox
                bboxes_data.append(
                    BboxData(
                        image_path=image_data.image_path,
                        xmin=xmin,
                        ymin=ymin,
                        xmax=xmax,
                        ymax=ymax,
                        keypoints=pred_keypoints,
                        mask=pred_mask,
                        detection_score=pred_detection_score,
                        label=pred_label_top_n[0],
                        classification_score=pred_classification_score_top_n[0],
                        top_n=len(pred_label_top_n),
                        labels_top_n=pred_label_top_n,
                        classification_scores_top_n=pred_classification_score_top_n,
                    )
                )
            if open_cropped_images_in_bboxes_data:
                for bbox_data in bboxes_data:
                    bbox_data.open_cropped_image(source_image=image_data.image, inplace=True)
            image = image_data.image if open_images_in_images_data else None
            pred_images_data.append(
                ImageData(
                    image_path=image_data.image_path,
                    image=image,
                    bboxes_data=bboxes_data,
                    label=image_data.label,
                    keypoints=image_data.keypoints,
                    mask=image_data.mask,
                    additional_info=image_data.additional_info,
                    meta_width=image_data.meta_width,
                    meta_height=image_data.meta_height,
                )
            )
        return pred_images_data

    def predict(
        self,
        images_data_gen: Union[List[ImageData], BatchGeneratorImageData],
        detection_score_threshold: float,
        classification_top_n: int = 1,
        open_images_in_images_data: bool = False,  # Warning: hard memory use
        open_cropped_images_in_bboxes_data: bool = False,
        disable_tqdm: bool = False,
        disable_tqdm_classification: bool = False,
        classification_batch_size: int = 16,
        detection_kwargs: Dict[str, Any] = {},
        classification_kwargs: Dict[str, Any] = {},
        progress_callback: Callable[[int], None] = lambda progress: None,
        batch_size_default: int = 16,
    ) -> List[ImageData]:
        if isinstance(images_data_gen, list):
            images_data_gen = BatchGeneratorImageData(images_data_gen, batch_size=batch_size_default)
        assert isinstance(images_data_gen, BatchGeneratorImageData)
        pred_images_data = []
        with tqdm(total=len(images_data_gen.data), disable=disable_tqdm) as pbar:
            for images_data in images_data_gen:
                input = [image_data.image for image_data in images_data]
                (
                    n_pred_bboxes,
                    n_k_pred_keypoints,
                    n_k_pred_masks,
                    n_pred_detection_scores,
                    n_pred_labels_top_n,
                    n_pred_classification_scores_top_n,
                ) = self.model.predict(
                    input=input,
                    detection_score_threshold=detection_score_threshold,
                    classification_top_n=classification_top_n,
                    classification_batch_size=classification_batch_size,
                    detection_kwargs=detection_kwargs,
                    classification_kwargs=classification_kwargs,
                    disable_tqdm_classification=disable_tqdm_classification,
                )
                pred_images_data_batch = self._postprocess_predictions(
                    images_data=images_data,
                    n_pred_bboxes=n_pred_bboxes,
                    n_k_pred_keypoints=n_k_pred_keypoints,
                    n_k_pred_masks=n_k_pred_masks,
                    n_pred_detection_scores=n_pred_detection_scores,
                    n_pred_labels_top_n=n_pred_labels_top_n,
                    n_pred_classification_scores_top_n=n_pred_classification_scores_top_n,
                    open_images_in_images_data=open_images_in_images_data,
                    open_cropped_images_in_bboxes_data=open_cropped_images_in_bboxes_data,
                )
                pred_images_data.extend(pred_images_data_batch)
                pbar.update(len(images_data))
                progress_callback(pbar.n)

        return pred_images_data

    @property
    def class_names(self):
        return self.model.class_names



from typing import Any, Dict, List, Tuple, Type, Union

import numpy as np
from tqdm import tqdm

from cv_pipeliner.core.inference_model import InferenceModel, ModelSpec
from cv_pipeliner.inference_models.classification.core import (
    ClassificationModel,
    ClassificationModelSpec,
)
from cv_pipeliner.inference_models.detection.core import (
    DetectionModel,
    DetectionModelSpec,
)
from cv_pipeliner.logging import logger
from cv_pipeliner.utils.images import cut_bboxes_from_image


class PipelineModelSpec(ModelSpec):
    detection_model_spec: DetectionModelSpec
    classification_model_spec: Union[ClassificationModelSpec, None] = None

    @property
    def inference_model_cls(self) -> Type["PipelineModel"]:
        from cv_pipeliner.inference_models.pipeline import PipelineModel

        return PipelineModel

    def load_pipeline_inferencer(self) -> "PipelineInferencer":
        from cv_pipeliner.inferencers.pipeline import PipelineInferencer

        return PipelineInferencer(self.load())


Bbox = Tuple[int, int, int, int]  # (ymin, xmin, ymax, xmax)
Score = float
Label = str
Keypoints = List[List[Tuple[int, int]]]
Mask = List[List[List[Tuple[int, int]]]]

Bboxes = List[Bbox]
DetectionScores = List[Score]
Labels = List[Label]
ClassificationScores = List[Score]

PipelineInput = List[np.ndarray]
PipelineOutput = List[
    Tuple[List[Bboxes], List[Keypoints], List[Mask], List[DetectionScores], List[Labels], List[ClassificationScores]]
]


class PipelineModel(InferenceModel):
    def __init__(self, model_spec: PipelineModelSpec = None):
        if model_spec is not None:
            isinstance(model_spec, PipelineModelSpec)
            super().__init__(model_spec)
            self.detection_model = model_spec.detection_model_spec.load()
            if model_spec.classification_model_spec is not None:
                self.classification_model = model_spec.classification_model_spec.load()
            else:
                self.classification_model = None

    def load_from_loaded_models(self, detection_model: DetectionModel, classification_model: ClassificationModel):
        isinstance(detection_model, DetectionModel)
        isinstance(classification_model, ClassificationModel)
        self._model_spec = PipelineModelSpec(
            detection_model_spec=detection_model.model_spec, classification_model_spec=classification_model.model_spec
        )
        self.detection_model = detection_model
        self.classification_model = classification_model

    def _split_chunks(self, _list: List, shapes: List[int]) -> List:
        cnt = 0
        chunks = []
        for shape in shapes:
            chunks.append(_list[cnt : cnt + shape])
            cnt += shape
        return chunks

    def predict(
        self,
        input: PipelineInput,
        detection_score_threshold: float,
        classification_top_n: int = 1,
        classification_batch_size: int = 16,
        detection_kwargs: Dict[str, Any] = {},
        classification_kwargs: Dict[str, Any] = {},
        disable_tqdm_classification: bool = False,
    ) -> PipelineOutput:
        logger.debug("Running detection...")
        (
            n_pred_bboxes,
            n_pred_keypoints,
            n_pred_masks,
            n_pred_detection_scores,
            n_pred_class_names_top_k,
            n_pred_classification_scores_top_k,
        ) = self.detection_model.predict(
            input,
            score_threshold=detection_score_threshold,
            classification_top_n=classification_top_n,
            **detection_kwargs,
        )
        logger.debug(f"Detection: found {np.sum([len(pred_bboxes) for pred_bboxes in n_pred_bboxes])} bboxes!")

        logger.debug("Running classification...")
        if self.classification_model is None:
            # Detector is the pipeline itself
            return (
                n_pred_bboxes,
                n_pred_keypoints,
                n_pred_masks,
                n_pred_detection_scores,
                n_pred_class_names_top_k,
                n_pred_classification_scores_top_k,
            )

        shapes = [len(pred_bboxes) for pred_bboxes in n_pred_bboxes]
        pred_labels_top_n, pred_classification_scores_top_n = [], []
        with tqdm(total=np.sum(shapes), disable=disable_tqdm_classification) as pbar:
            for image, pred_bboxes in zip(input, n_pred_bboxes):
                if len(pred_bboxes) == 0:
                    continue
                pred_bboxes_batches = np.array_split(pred_bboxes, max(1, len(pred_bboxes) // classification_batch_size))
                for pred_bboxes_batch in pred_bboxes_batches:
                    classification_input_batch = cut_bboxes_from_image(image, pred_bboxes_batch)
                    pred_labels_top_n_batch, pred_classification_scores_top_n_batch = self.classification_model.predict(
                        input=classification_input_batch, top_n=classification_top_n, **classification_kwargs
                    )
                    pred_labels_top_n.extend(pred_labels_top_n_batch)
                    pred_classification_scores_top_n.extend(pred_classification_scores_top_n_batch)
                    pbar.update(len(pred_bboxes_batch))
        n_pred_labels_top_n = self._split_chunks(pred_labels_top_n, shapes)
        n_pred_classification_scores_top_n = self._split_chunks(pred_classification_scores_top_n, shapes)
        logger.debug("Classification end!")
        return (
            n_pred_bboxes,
            n_pred_keypoints,
            n_pred_masks,
            n_pred_detection_scores,
            n_pred_labels_top_n,
            n_pred_classification_scores_top_n,
        )

    def preprocess_input(self, input):
        return input

    @property
    def input_size(self) -> Tuple[int, int]:
        return self.detection_model.input_size

    @property
    def class_names(self) -> List[str]:
        if self.classification_model is not None:
            return self.classification_model.class_names
        elif self.detection_model.class_names is not None:
            return self.detection_model.class_names



import abc
from typing import List, Tuple, Type

import numpy as np

from cv_pipeliner.core.inference_model import InferenceModel, ModelSpec

Keypoint = Tuple[int, int]
Keypoints = List[Keypoint]

KeypointsRegressorInput = List[np.ndarray]
KeypointsRegressorOutput = List[Keypoints]


class KeypointsRegressorModelSpec(ModelSpec):
    @abc.abstractproperty
    def inference_model_cls(self) -> Type["KeypointsRegressorModel"]:
        pass

    def load_keypoints_regressor_inferencer(self) -> "KeypointsRegressorInferencer":
        from cv_pipeliner.inferencers.keypoints_regressor import (
            KeypointsRegressorInferencer,
        )

        return KeypointsRegressorInferencer(self.load())


class KeypointsRegressorModel(InferenceModel):
    def __init__(self, model_spec: KeypointsRegressorModelSpec):
        assert isinstance(model_spec, KeypointsRegressorModelSpec)
        super().__init__(model_spec)

    @abc.abstractmethod
    def predict(
        self,
        input: KeypointsRegressorInput,
    ) -> KeypointsRegressorOutput:
        pass

    @abc.abstractmethod
    def preprocess_input(self, input):
        pass

    @abc.abstractproperty
    def input_size(self) -> Tuple[int, int]:
        pass



import logging
from io import BytesIO
from pathlib import Path
from typing import Dict, List, Tuple, Union

import contextlib2
import numpy as np
import tensorflow as tf
from google.protobuf import text_format
from joblib import Parallel, delayed
from object_detection.dataset_tools import tf_record_creation_util as creation_util
from object_detection.protos import pipeline_pb2
from object_detection.protos.string_int_label_map_pb2 import (
    StringIntLabelMap,
    StringIntLabelMapItem,
)
from object_detection.utils import dataset_util
from object_detection.utils.config_util import (
    create_pipeline_proto_from_configs,
    get_configs_from_pipeline_file,
    save_pipeline_config,
)
from PIL import Image
from tqdm import tqdm

from cv_pipeliner.core.data import ImageData

logger = logging.getLogger(__name__)


def create_tf_record(
    height: int,
    width: int,
    encoded_filename: bytes,
    encoded_jpg: bytes,
    image_format: bytes,
    xmins: List[float],
    ymins: List[float],
    xmaxs: List[float],
    ymaxs: List[float],
    encoded_class_names: List[bytes],
    classes: int,
) -> tf.train.Example:
    tf_record = tf.train.Example(
        features=tf.train.Features(
            feature={
                "image/height": dataset_util.int64_feature(height),
                "image/width": dataset_util.int64_feature(width),
                "image/filename": dataset_util.bytes_feature(encoded_filename),
                "image/source_id": dataset_util.bytes_feature(encoded_filename),
                "image/encoded": dataset_util.bytes_feature(encoded_jpg),
                "image/format": dataset_util.bytes_feature(image_format),
                "image/object/bbox/xmin": dataset_util.float_list_feature(xmins),
                "image/object/bbox/ymin": dataset_util.float_list_feature(ymins),
                "image/object/bbox/xmax": dataset_util.float_list_feature(xmaxs),
                "image/object/bbox/ymax": dataset_util.float_list_feature(ymaxs),
                "image/object/class/text": dataset_util.bytes_list_feature(encoded_class_names),
                "image/object/class/label": dataset_util.int64_list_feature(classes),
            }
        )
    )
    return tf_record


def tf_record_from_image_data(image_data: ImageData, label_map: Dict[str, int], use_thumbnail: Tuple[int, int] = None):
    filename = image_data.image_path
    encoded_filename = str(filename).encode("utf8")

    normalized_true_bboxes = np.array([bbox_data.coords_n for bbox_data in image_data.bboxes_data], dtype=float)
    image = image_data.open_image()
    width, height = image_data.get_image_size()
    if len(normalized_true_bboxes) > 0:
        xmins = normalized_true_bboxes[:, 0]
        ymins = normalized_true_bboxes[:, 1]
        xmaxs = normalized_true_bboxes[:, 2]
        ymaxs = normalized_true_bboxes[:, 3]
    else:
        ymins, xmins, ymaxs, xmaxs = [], [], [], []

    encoded_jpg = BytesIO()
    image = Image.fromarray(image)
    if use_thumbnail:
        image.thumbnail(use_thumbnail)
    image.save(encoded_jpg, format="JPEG")
    encoded_jpg = encoded_jpg.getvalue()
    image_format = b"jpg"

    class_names = [bbox_data.label for bbox_data in image_data.bboxes_data]
    encoded_class_names = [class_name.encode("utf-8") for class_name in class_names]
    classes = [label_map[class_name] for class_name in class_names]

    tf_record = create_tf_record(
        height=height,
        width=width,
        encoded_filename=encoded_filename,
        encoded_jpg=encoded_jpg,
        image_format=image_format,
        xmins=xmins,
        ymins=ymins,
        xmaxs=xmaxs,
        ymaxs=ymaxs,
        encoded_class_names=encoded_class_names,
        classes=classes,
    )
    return tf_record


def convert_to_tf_records(
    images_data: List[ImageData],
    label_map: Dict[str, int],
    filepath: Union[str, Path],
    num_workers: int = 1,
    num_shards: int = 2,
    max_pictures_per_worker: int = 1000,
    use_thumbnail: Tuple[int, int] = None,
):
    logger.info("Create tf_records from data.")

    data_chunks = np.array_split(images_data, max_pictures_per_worker)
    with contextlib2.ExitStack() as tf_record_close_stack:
        output_tfrecords = creation_util.open_sharded_output_tfrecords(
            exit_stack=tf_record_close_stack, base_path=filepath, num_shards=num_shards
        )
        for data_chunk in tqdm(data_chunks):
            tf_records = Parallel(n_jobs=num_workers, prefer="threads")(
                delayed(tf_record_from_image_data)(
                    image_data=image_data, label_map=label_map, use_thumbnail=use_thumbnail
                )
                for image_data in data_chunk
            )

            for index, tf_record in enumerate(tf_records):
                output_shard_index = index % num_shards
                output_tfrecords[output_shard_index].write(tf_record.SerializeToString())

    logger.info(f"tf_records saved to {filepath}-?????-of-{str(num_shards).zfill(5)}.")


def label_map_to_file(label_map: Dict[str, int], filepath: Union[str, Path]):
    msg = StringIntLabelMap()
    label_map = {label: i for label, i in sorted(label_map.items(), key=lambda item: item[1])}
    for label, i in label_map.items():
        # pylint: disable=no-member
        msg.item.append(StringIntLabelMapItem(id=i, name=label))

    text = str(text_format.MessageToBytes(msg, as_utf8=True), "utf-8")
    with open(filepath, "w") as out:
        out.write(text)
    logger.info(f"label_map saved to {filepath}")


def count_tfrecord_examples(filepath: Union[str, Path]):
    filepath = Path(filepath)
    if filepath.exists():
        count = sum(1 for _ in tf.data.TFRecordDataset(str(filepath)))
    else:
        filepaths = sorted(filepath.parent.glob(f"{filepath.name}*-of-*"))
        count = sum(1 for filepath in filepaths for _ in tf.data.TFRecordDataset(str(filepath)))
    return count


def set_config(
    config_path: Union[str, Path],
    checkpoint_path: Union[str, Path],
    tf_records_train_path: Union[str, Path],
    label_map: Dict[str, int],
    label_map_filepath: Union[str, Path],
    batch_size: int,
    max_box_predictions: int,
    max_number_of_boxes: int,
    fine_tune_checkpoint_type: str = "detection",
    augment_path: str = None,
    min_dimension: int = None,
    max_dimension: int = None,
    total_steps: int = None,
    warmup_steps: int = None,
    num_steps: int = None,
):
    logger.info(f"Set configs {config_path}...")

    configs = get_configs_from_pipeline_file(str(config_path))

    train_len = count_tfrecord_examples(str(tf_records_train_path))
    logger.info(f"Train has {train_len} tf_records.")
    num_classes = len(set(label_map.values()))
    _, config_model = configs["model"].ListFields()[0]
    config_model.num_classes = num_classes

    configs["model"].center_net.object_center_params.max_box_predictions = max_box_predictions
    if min_dimension is not None:
        configs["model"].center_net.image_resizer.keep_aspect_ratio_resizer.min_dimension = min_dimension
    if max_dimension is not None:
        configs["model"].center_net.image_resizer.keep_aspect_ratio_resizer.max_dimension = max_dimension

    configs["train_config"].fine_tune_checkpoint_type = fine_tune_checkpoint_type
    configs["train_config"].fine_tune_checkpoint = str(checkpoint_path)
    configs["train_config"].batch_size = batch_size

    configs["train_config"].max_number_of_boxes = max_number_of_boxes
    if total_steps is not None:
        configs["train_config"].optimizer.adam_optimizer.learning_rate.cosine_decay_learning_rate.total_steps = (
            total_steps
        )
    if warmup_steps is not None:
        configs["train_config"].optimizer.adam_optimizer.learning_rate.cosine_decay_learning_rate.warmup_steps = (
            warmup_steps
        )
    if num_steps is not None:
        configs["train_config"].num_steps = num_steps

    if augment_path is not None:
        augment_config = configs["train_config"].data_augmentation_options
        for _ in augment_config:
            augment_config.pop()
        augment = text_format.Merge(str(augment_path), pipeline_pb2.TrainEvalPipelineConfig())
        augment_config.extend(augment.train_config.data_augmentation_options)

    label_map_to_file(label_map=label_map, filepath=label_map_filepath)

    def clear_repeated_proto(proto):
        for _ in proto:
            proto.pop()

    configs["train_input_config"].label_map_path = str(label_map_filepath)
    clear_repeated_proto(configs["train_input_config"].tf_record_input_reader.input_path)
    configs["train_input_config"].tf_record_input_reader.input_path.append(str(tf_records_train_path))

    pipeline_proto = create_pipeline_proto_from_configs(configs)
    save_pipeline_config(pipeline_proto, str(config_path.parent))
    logger.info(f"Config {config_path} changed")



import numpy as np


def voc_bboxes_to_coco(bboxes):
    bboxes = np.array(bboxes)
    if len(bboxes) > 0:
        bboxes[:, 2] = bboxes[:, 2] - bboxes[:, 0]
        bboxes[:, 3] = bboxes[:, 3] - bboxes[:, 1]
    result = list(map(tuple, bboxes))
    return result


def coco_bboxes_to_voc(bboxes, frame_size):
    bboxes = np.array(bboxes)
    if len(bboxes) > 0:
        bboxes = np.clip(bboxes, 0, max(frame_size))
        bboxes = bboxes.round().astype("int")
        bboxes[:, 2] = bboxes[:, 0] + bboxes[:, 2]
        bboxes[:, 3] = bboxes[:, 1] + bboxes[:, 3]
    return bboxes



# import json
# import requests
# from pathlib import Path

# examples_dir = Path(__file__).absolute().parent.parent / 'examples/brickit-ml/'
# tmp_dir = Path('tmp/')
# tmp_dir.mkdir(exist_ok=True)

# example_file = examples_dir / '35f20998-47ff-4d74-b2d9-245edc3dec32.jpeg'
# files = {'image': open(example_file, 'rb')}
# print(f'POST for {example_file}')
# response = requests.post("http://localhost:5000/predict", files=files)
# with open(tmp_dir / f'{example_file.stem}.json', 'w') as out:
#     json.dump(json.loads(response.text), out, indent=4)

# guid = 'AABBCC'
# response = requests.post(
#     f"http://localhost:5000/realtime_start/{guid}",
#     data={
#         'fps': 30,
#         'detection_delay': 300
#     }
# )
# frames_files = (examples_dir / 'easy').glob('*.png')
# for frame_file in frames_files:
#     print(f'POST for {frame_file}...')
#     response = requests.post(
#         f"http://localhost:5000/realtime_predict/{guid}",
#         files={'image': open(frame_file, 'rb')}
#     )
#     with open(tmp_dir / f'{frame_file.stem}.json', 'w') as out:
#         json.dump(json.loads(response.text), out, indent=4)
# response = requests.post(f"http://localhost:5000/realtime_end/{guid}")



from typing import List

import numpy as np

from cv_pipeliner.core.batch_generator import BatchGenerator
from cv_pipeliner.core.data import BboxData


class BatchGeneratorBboxData(BatchGenerator):
    def __init__(
        self,
        data: List[List[BboxData]],
        batch_size: int,
        use_not_caught_elements_as_last_batch: bool = True,
        open_cropped_images: bool = True,
    ):
        assert all(isinstance(d, list) or isinstance(d, np.ndarray) for d in data)
        assert all(isinstance(item, BboxData) for d in data for item in d)
        self._shapes = np.array([len(subdata) for subdata in data])
        data = [item for sublist in data for item in sublist]
        super().__init__(data, batch_size, use_not_caught_elements_as_last_batch)
        self.open_cropped_images = open_cropped_images

    def __getitem__(self, index: int) -> List[BboxData]:
        batch = super().__getitem__(index)
        unique_image_paths, unique_image_idxs = np.unique(
            [bbox_data.image_path for bbox_data in batch if bbox_data.image_path is not None], return_index=True
        )
        unique_image_idx_to_image = {
            str(unique_image_path): batch[unique_image_idx].open_image()
            for unique_image_path, unique_image_idx in zip(unique_image_paths, unique_image_idxs)
        }
        for bbox_data in batch:
            source_image = (
                unique_image_idx_to_image[str(bbox_data.image_path)] if bbox_data.image_path is not None else None
            )
            if self.open_cropped_images and bbox_data.cropped_image is None:
                bbox_data.open_cropped_image(source_image=source_image, inplace=True)
        return batch

    @property
    def shapes(self):
        return self._shapes

    def __iter__(self):
        for bboxes_data in (self[i] for i in range(len(self))):
            yield bboxes_data
            for bbox_data in bboxes_data:
                bbox_data.image = None
                bbox_data.cropped_image = None
            del bboxes_data



import copy
from typing import List, Literal, Optional, Tuple, Union

import cv2
import numpy as np

from cv_pipeliner.core.data import BboxData, ImageData
from cv_pipeliner.utils.images import concat_images, get_thumbnail_resize


def get_image_data_filtered_by_labels(
    image_data: ImageData, filter_by_labels: List[str] = None, include: bool = True
) -> ImageData:
    if filter_by_labels is None or len(filter_by_labels) == 0:
        return image_data
    image_data = copy.deepcopy(image_data)

    filter_by_labels = set(filter_by_labels)

    image_data.bboxes_data = [
        bbox_data
        for bbox_data in image_data.bboxes_data
        if (include and bbox_data.label in filter_by_labels)
        or (not include and bbox_data.label not in filter_by_labels)
    ]
    return image_data


def get_n_bboxes_data_filtered_by_labels(
    n_bboxes_data: List[List[BboxData]], filter_by_labels: List[str] = None, include: bool = True
) -> ImageData:
    if filter_by_labels is None or len(filter_by_labels) == 0:
        return n_bboxes_data

    filter_by_labels = set(filter_by_labels)

    n_bboxes_data = [
        [
            bbox_data
            for bbox_data in bboxes_data
            if (include and bbox_data.label in filter_by_labels)
            or (not include and bbox_data.label not in filter_by_labels)
        ]
        for bboxes_data in n_bboxes_data
    ]
    return n_bboxes_data


def rotate_keypoints(keypoints: Tuple[Tuple[int, int]], rotation_mat: np.ndarray, new_width: int, new_height: int):
    keypoints = np.array(keypoints)
    points = np.zeros((len(keypoints), 3))
    points[:, 0] = keypoints[:, 0]
    points[:, 1] = keypoints[:, 1]
    points[:, 2] = 1
    rotated_points = (rotation_mat @ points.T).astype(int).T
    rotated_points = np.array(rotated_points).reshape(-1, 2)
    rotated_points[:, 0] = np.clip(rotated_points[:, 0], 0, new_width - 1)
    rotated_points[:, 1] = np.clip(rotated_points[:, 1], 0, new_height - 1)
    return rotated_points


def rotate_keypoints90(
    keypoints: Tuple[Tuple[int, int]],
    factor: Literal[0, 1, 2, 3],
    width: int,
    height: int,
) -> np.ndarray:
    """Rotates a bounding box by 90 degrees CCW (see np.rot90)"""
    rotated_keypoints = []
    for x, y in keypoints:
        if factor == 1:
            x, y = y, (width - 1) - x
        elif factor == 2:
            x, y = (width - 1) - x, (height - 1) - y
        elif factor == 3:
            x, y = (height - 1) - y, x
        rotated_keypoints.append([x, y])
    return np.array(rotated_keypoints).reshape(-1, 2)


def _rotate_bbox_data(bbox_data: BboxData, rotation_mat: np.ndarray, new_width: int, new_height: int) -> BboxData:
    bbox_points = np.array(
        [
            [bbox_data.xmin, bbox_data.ymin],
            [bbox_data.xmin, bbox_data.ymax],
            [bbox_data.xmax, bbox_data.ymin],
            [bbox_data.xmax, bbox_data.ymax],
        ]
    )
    rotated_points = rotate_keypoints(bbox_points, rotation_mat, new_width, new_height)
    rotated_xmin = max(0, min(np.min(rotated_points[:, 0]), new_width - 1))
    rotated_ymin = max(0, min(np.min(rotated_points[:, 1]), new_height - 1))
    rotated_xmax = max(0, min(np.max(rotated_points[:, 0]), new_width - 1))
    rotated_ymax = max(0, min(np.max(rotated_points[:, 1]), new_height - 1))
    rotated_bbox_data = copy.deepcopy(bbox_data)
    rotated_bbox_data.xmin = rotated_xmin
    rotated_bbox_data.ymin = rotated_ymin
    rotated_bbox_data.xmax = rotated_xmax
    rotated_bbox_data.ymax = rotated_ymax
    rotated_bbox_data.keypoints = rotate_keypoints(rotated_bbox_data.keypoints, rotation_mat, new_width, new_height)
    if isinstance(rotated_bbox_data.mask, list):
        rotated_bbox_data.mask = [
            rotate_keypoints(polygon, rotation_mat, new_width, new_height) for polygon in rotated_bbox_data.mask
        ]
        for polygon in rotated_bbox_data.mask:
            polygon[:, 0] = np.clip(polygon[:, 0], 0, new_width - 1)
            polygon[:, 1] = np.clip(polygon[:, 1], 0, new_height - 1)
    keypoints = []
    for x, y in rotated_bbox_data.keypoints:
        x = max(0, min(x, new_width - 1))
        y = max(0, min(y, new_height - 1))
        keypoints.append([x, y])
    rotated_bbox_data.keypoints = np.array(keypoints).reshape(-1, 2)
    rotated_bbox_data.additional_bboxes_data = [
        _rotate_bbox_data(additional_bbox_data, rotation_mat, new_width, new_height)
        for additional_bbox_data in rotated_bbox_data.additional_bboxes_data
    ]
    rotated_bbox_data.cropped_image = None
    return rotated_bbox_data


def _rotate_bbox_data90(
    bbox_data: BboxData,
    factor: Literal[0, 1, 2, 3],
    width: int,
    height: int,
) -> BboxData:
    """Rotates a bounding box by 90 degrees CCW (see np.rot90)"""
    rotated_bbox_data = copy.deepcopy(bbox_data)
    rotated_bbox_data.keypoints = rotate_keypoints90(bbox_data.keypoints, factor, width, height)
    if isinstance(rotated_bbox_data.mask, list):
        rotated_bbox_data.mask = [
            rotate_keypoints90(polygon, factor, width, height) for polygon in rotated_bbox_data.mask
        ]
    xmin, ymin, xmax, ymax = rotated_bbox_data.coords
    if factor == 1:
        xmin, ymin, xmax, ymax = ymin, width - xmax, ymax, width - xmin
    elif factor == 2:
        xmin, ymin, xmax, ymax = width - xmax, height - ymax, width - xmin, height - ymin
    elif factor == 3:
        xmin, ymin, xmax, ymax = height - ymax, xmin, height - ymin, xmax
    rotated_bbox_data.xmin = xmin
    rotated_bbox_data.ymin = ymin
    rotated_bbox_data.xmax = xmax
    rotated_bbox_data.ymax = ymax
    rotated_bbox_data.additional_bboxes_data = [
        _rotate_bbox_data90(additional_bbox_data, factor, width, height)
        for additional_bbox_data in rotated_bbox_data.additional_bboxes_data
    ]
    rotated_bbox_data.cropped_image = None
    return rotated_bbox_data


def rotate_image_data(
    image_data: ImageData,
    angle: float,
    warp_flags: Optional[int] = None,
    border_mode: Optional[int] = None,
    border_value: Tuple[int, int, int] = None,
    open_image: bool = True,
    exif_transpose: bool = False,
):
    rotated_image_data = copy.deepcopy(image_data)
    if abs(angle) <= 1e-6:
        if open_image:
            rotated_image_data.open_image(inplace=True, exif_transpose=exif_transpose)
        return rotated_image_data

    width, height = image_data.get_image_size()
    image = image_data.open_image(returns_none_if_empty=True, exif_transpose=exif_transpose) if open_image else None
    image_center = width // 2, height // 2

    angle_to_factor = {0: 0, 90: 1, 180: 2, 270: 3}
    angle = angle % 360

    if angle in angle_to_factor:
        factor = angle_to_factor[angle]
        rotated_image = np.rot90(image, factor) if image is not None else None
        rotated_image_data.keypoints = rotate_keypoints90(image_data.keypoints, factor, width, height)
        if isinstance(rotated_image_data.mask, list):
            rotated_image_data.keypoints = [
                rotate_keypoints90(polygon, factor, width, height) for polygon in rotated_image_data.mask
            ]
        rotated_image_data.bboxes_data = [
            _rotate_bbox_data90(bbox_data, factor, width, height) for bbox_data in rotated_image_data.bboxes_data
        ]
        new_width, new_height = (width, height) if (factor % 2 == 0) else (height, width)
    else:
        # grab the rotation matrix
        rotation_mat = cv2.getRotationMatrix2D(image_center, angle, 1.0)
        # compute the new bounding dimensions of the image
        abs_cos = abs(rotation_mat[0, 0])
        abs_sin = abs(rotation_mat[0, 1])
        new_width = int(height * abs_sin + width * abs_cos)
        new_height = int(height * abs_cos + width * abs_sin)
        # adjust the rotation matrix to take into account translation
        rotation_mat[0, 2] += new_width / 2 - image_center[0]
        rotation_mat[1, 2] += new_height / 2 - image_center[1]

        rotated_image = (
            cv2.warpAffine(
                image,
                rotation_mat,
                (new_width, new_height),
                flags=warp_flags,
                borderMode=border_mode,
                borderValue=border_value,
            )
            if image is not None
            else None
        )
        rotated_image_data = copy.deepcopy(image_data)
        rotated_image_data.keypoints = rotate_keypoints(image_data.keypoints, rotation_mat, new_height, new_width)
        if isinstance(rotated_image_data.mask, list):
            rotated_image_data.keypoints = [
                rotate_keypoints(polygon, rotation_mat, new_height, new_width) for polygon in rotated_image_data.mask
            ]
        keypoints = []
        for x, y in rotated_image_data.keypoints:
            x = max(0, min(x, new_width - 1))
            y = max(0, min(y, new_height - 1))
            keypoints.append([x, y])
        rotated_image_data.keypoints = np.array(keypoints).reshape(-1, 2)
        if isinstance(rotated_image_data.mask, list):
            for polygon in rotated_image_data.mask:
                polygon[:, 0] = np.clip(polygon[:, 0], 0, new_width - 1)
                polygon[:, 1] = np.clip(polygon[:, 1], 0, new_height - 1)
        rotated_image_data.bboxes_data = [
            _rotate_bbox_data(bbox_data, rotation_mat, new_height, new_width)
            for bbox_data in rotated_image_data.bboxes_data
        ]

    rotated_image_data.image_path = None  # It applies to all bboxes_data inside
    rotated_image_data.meta_height = new_height
    rotated_image_data.meta_width = new_width
    rotated_image_data.image = rotated_image

    return rotated_image_data


def resize_image_data(
    image_data: ImageData,
    size: Tuple[int, int],
    interpolation: Optional[int] = cv2.INTER_LINEAR,
    open_image: bool = True,
    exif_transpose: bool = False,
) -> ImageData:
    image_data = copy.deepcopy(image_data)
    old_width, old_height = image_data.get_image_size()
    new_width, new_height = size

    image = image_data.open_image(returns_none_if_empty=True, exif_transpose=exif_transpose) if open_image else None
    image = cv2.resize(image, size, interpolation=interpolation) if image is not None else None

    def resize_coords(bbox_data: BboxData):
        bbox_data.xmin = max(0, min(int(bbox_data.xmin * (new_width / old_width)), new_width - 1))
        bbox_data.ymin = max(0, min(int(bbox_data.ymin * (new_height / old_height)), new_height - 1))
        bbox_data.xmax = max(0, min(int(bbox_data.xmax * (new_width / old_width)), new_width - 1))
        bbox_data.ymax = max(0, min(int(bbox_data.ymax * (new_height / old_height)), new_height - 1))
        bbox_data.keypoints[:, 0] = (bbox_data.keypoints[:, 0] * (new_width / old_width)).astype(int)
        bbox_data.keypoints[:, 1] = (bbox_data.keypoints[:, 1] * (new_height / old_height)).astype(int)
        if isinstance(bbox_data.mask, list):
            for polygon in bbox_data.mask:
                polygon[:, 0] = (polygon[:, 0] * (new_width / old_width)).astype(int)
                polygon[:, 1] = (polygon[:, 1] * (new_height / old_height)).astype(int)
                polygon[:, 0] = np.clip(polygon[:, 0], 0, new_width - 1)
                polygon[:, 1] = np.clip(polygon[:, 1], 0, new_height - 1)
        bbox_data.keypoints = bbox_data.keypoints.astype(int)
        bbox_data.cropped_image = None
        keypoints = []
        for x, y in bbox_data.keypoints:
            x = max(0, min(x, new_width - 1))
            y = max(0, min(y, new_height - 1))
            keypoints.append([x, y])
        bbox_data.keypoints = np.array(keypoints).reshape(-1, 2)
        for additional_bbox_data in bbox_data.additional_bboxes_data:
            resize_coords(additional_bbox_data)

    for bbox_data in image_data.bboxes_data:
        resize_coords(bbox_data)
    image_data.keypoints[:, 0] = (image_data.keypoints[:, 0] * (new_width / old_width)).astype(int)
    image_data.keypoints[:, 1] = (image_data.keypoints[:, 1] * (new_height / old_height)).astype(int)
    keypoints = []
    for x, y in image_data.keypoints:
        x = max(0, min(x, new_width - 1))
        y = max(0, min(y, new_height - 1))
        keypoints.append([x, y])
    if isinstance(image_data.mask, list):
        for polygon in image_data.mask:
            polygon[:, 0] = (polygon[:, 0] * (new_width / old_width)).astype(int)
            polygon[:, 1] = (polygon[:, 1] * (new_height / old_height)).astype(int)
            polygon[:, 0] = np.clip(polygon[:, 0], 0, new_width - 1)
            polygon[:, 1] = np.clip(polygon[:, 1], 0, new_height - 1)
    image_data.keypoints = np.array(keypoints).reshape(-1, 2)
    image_data.image_path = None
    image_data.image = image

    return image_data


def thumbnail_image_data(
    image_data: ImageData,
    size: Optional[Union[int, Tuple[int, int]]] = None,
    interpolation: Optional[int] = cv2.INTER_LINEAR,
) -> ImageData:
    if isinstance(size, int):
        size = (size, size)
    new_width, new_height = get_thumbnail_resize(image_data.get_image_size(), size)
    return resize_image_data(image_data, (new_width, new_height), interpolation=interpolation)


def crop_image_data(
    image_data: ImageData,
    xmin: int,
    ymin: int,
    xmax: int,
    ymax: int,
    allow_negative_and_large_coords: bool,
    remove_bad_coords: bool,
    open_image: bool = True,
    exif_transpose: bool = False,
) -> ImageData:
    assert 0 <= xmin and 0 <= ymin
    assert xmin <= xmax and ymin <= ymax

    image_data = copy.deepcopy(image_data)
    image = image_data.open_image(returns_none_if_empty=True, exif_transpose=exif_transpose) if open_image else None
    width, height = image_data.get_image_size()
    xmin = max(0, xmin)
    ymin = max(0, ymin)
    xmax = min(xmax, width - 1)
    ymax = min(ymax, height - 1)

    assert (
        xmin >= 0 and ymin >= 0 and xmin < xmax and ymin < ymax and xmax <= width - 1 and ymax <= height - 1
    ), f"Wrong arguments: {(xmin, ymin, xmax, ymax)=} ({width=}, {height=})"

    image = image[ymin:ymax, xmin:xmax] if image is not None else None
    new_width, new_height = max(0, min(width - 1, xmax - xmin)), max(0, min(height - 1, ymax - ymin))

    def resize_coords(bbox_data: BboxData):
        bbox_data.xmin = bbox_data.xmin - xmin
        bbox_data.ymin = bbox_data.ymin - ymin
        bbox_data.xmax = bbox_data.xmax - xmin
        bbox_data.ymax = bbox_data.ymax - ymin
        bbox_data.keypoints[:, 0] -= xmin
        bbox_data.keypoints[:, 1] -= ymin
        if isinstance(bbox_data.mask, list):
            for polygon in bbox_data.mask:
                polygon[:, 0] -= xmin
                polygon[:, 1] -= ymin
        bbox_data.cropped_image = None
        if not allow_negative_and_large_coords:
            bbox_data.xmin = max(0, min(bbox_data.xmin, new_width - 1))
            bbox_data.ymin = max(0, min(bbox_data.ymin, new_height - 1))
            bbox_data.xmax = max(0, min(bbox_data.xmax, new_width - 1))
            bbox_data.ymax = max(0, min(bbox_data.ymax, new_height - 1))
            keypoints = []
            for x, y in bbox_data.keypoints:
                x = max(0, min(x, new_width - 1))
                y = max(0, min(y, new_height - 1))
                keypoints.append([x, y])
            bbox_data.keypoints = np.array(keypoints).reshape(-1, 2)
            if isinstance(bbox_data.mask, list):
                for polygon in bbox_data.mask:
                    polygon[:, 0] = np.clip(polygon[:, 0], 0, new_width - 1)
                    polygon[:, 1] = np.clip(polygon[:, 1], 0, new_height - 1)
        for additional_bbox_data in bbox_data.additional_bboxes_data:
            resize_coords(additional_bbox_data)

    for bbox_data in image_data.bboxes_data:
        resize_coords(bbox_data)

    keypoints = []
    for x, y in image_data.keypoints:
        x = max(0, min(x - xmin, new_width - 1))
        y = max(0, min(y - ymin, new_height - 1))
        keypoints.append([x, y])
    image_data.keypoints = np.array(keypoints).reshape(-1, 2)
    if isinstance(image_data.mask, list):
        for polygon in image_data.mask:
            polygon[:, 0] = np.clip(polygon[:, 0], 0, new_width - 1)
            polygon[:, 1] = np.clip(polygon[:, 1], 0, new_height - 1)

    def if_bbox_data_inside_crop(bbox_data: BboxData):
        bbox_data.keypoints = bbox_data.keypoints[
            (
                (bbox_data.keypoints[:, 0] >= 0)
                & (bbox_data.keypoints[:, 1] >= 0)
                & (bbox_data.keypoints[:, 0] < new_height)
                & (bbox_data.keypoints[:, 1] < new_width)
            )
        ]
        if isinstance(bbox_data.mask, list):
            bbox_data.mask = [
                polygon[
                    (
                        (polygon[:, 0] >= 0)
                        & (polygon[:, 1] >= 0)
                        & (polygon[:, 0] < new_height)
                        & (polygon[:, 1] < new_width)
                    )
                ]
                for polygon in bbox_data.mask
            ]
            bbox_data.mask = [polygon for polygon in bbox_data.mask if len(polygon) > 0]
        bbox_data.additional_bboxes_data = [
            additional_bbox_data
            for additional_bbox_data in bbox_data.additional_bboxes_data
            if if_bbox_data_inside_crop(additional_bbox_data)
        ]
        return (
            bbox_data.xmin >= 0
            and bbox_data.ymin >= 0
            and bbox_data.xmax < new_width
            and bbox_data.ymax < new_height
            and bbox_data.xmin < bbox_data.xmax
            and bbox_data.ymin < bbox_data.ymax
        )

    if remove_bad_coords:
        image_data.bboxes_data = [
            bbox_data for bbox_data in image_data.bboxes_data if if_bbox_data_inside_crop(bbox_data)
        ]
        image_data.keypoints = image_data.keypoints[
            (
                (image_data.keypoints[:, 0] >= 0)
                & (image_data.keypoints[:, 1] >= 0)
                & (image_data.keypoints[:, 0] < new_height)
                & (image_data.keypoints[:, 1] < new_width)
            )
        ]
        if isinstance(image_data.mask, list):
            image_data.mask = [
                polygon[
                    (
                        (polygon[:, 0] >= 0)
                        & (polygon[:, 1] >= 0)
                        & (polygon[:, 0] < new_height)
                        & (polygon[:, 1] < new_width)
                    )
                ]
                for polygon in image_data.mask
            ]
            image_data.mask = [polygon for polygon in image_data.mask if len(polygon) > 0]

    image_data.image_path = None
    image_data.image = image

    return image_data


def apply_perspective_transform_to_points(
    points: List[Tuple[int, int]],
    perspective_matrix: np.ndarray,
    result_width: int,
    result_height: int,
    allow_negative_and_large_coords: bool,
    remove_bad_coords: bool,
):
    points = np.array(points)
    if len(points) == 0:
        return points
    transformed_points = (
        cv2.perspectiveTransform(points.reshape(1, -1, 2).astype(np.float32), perspective_matrix)
        .reshape(-1, 2)
        .astype(int)
    )
    if not allow_negative_and_large_coords:
        transformed_points_without_bad_coords = []
        for x, y in points:
            x = max(0, min(x, result_width - 1))
            y = max(0, min(y, result_height - 1))
            transformed_points_without_bad_coords.append([x, y])
        transformed_points = np.array(transformed_points_without_bad_coords)
    if remove_bad_coords:
        transformed_points = transformed_points[
            (transformed_points[:, 0] >= 0)
            & (transformed_points[:, 1] >= 0)
            & (transformed_points[:, 0] < result_width)
            & (transformed_points[:, 1] < result_height)
        ]
    return transformed_points


def _apply_perspective_transform_to_bbox_data(
    bbox_data: BboxData,
    perspective_matrix: np.ndarray,
    result_width: int,
    result_height: int,
    allow_negative_and_large_coords: bool,
    remove_bad_coords: bool,
) -> BboxData:
    bbox_points = np.array(
        [
            [bbox_data.xmin, bbox_data.ymin],
            [bbox_data.xmin, bbox_data.ymax],
            [bbox_data.xmax, bbox_data.ymin],
            [bbox_data.xmax, bbox_data.ymax],
        ]
    )
    transformed_points = (
        cv2.perspectiveTransform(bbox_points.reshape(1, -1, 2).astype(np.float32), perspective_matrix)
        .reshape(-1, 2)
        .astype(int)
    )
    transformed_xmin = np.min(transformed_points[:, 0])
    transformed_ymin = np.min(transformed_points[:, 1])
    transformed_xmax = np.max(transformed_points[:, 0])
    transformed_ymax = np.max(transformed_points[:, 1])
    if not allow_negative_and_large_coords:
        transformed_xmin = max(0, min(transformed_xmin, result_width - 1))
        transformed_ymin = max(0, min(transformed_ymin, result_height - 1))
        transformed_xmax = max(0, min(transformed_xmax, result_width - 1))
        transformed_ymax = max(0, min(transformed_ymax, result_height - 1))
    if remove_bad_coords and not (
        transformed_xmin >= 0
        and transformed_ymin >= 0
        and transformed_xmax < result_width
        and transformed_ymax < result_height
        and transformed_xmin < transformed_xmax
        and transformed_ymin < transformed_ymax
    ):
        return None

    transformed_bbox_data = copy.deepcopy(bbox_data)
    transformed_bbox_data.xmin = transformed_xmin
    transformed_bbox_data.ymin = transformed_ymin
    transformed_bbox_data.xmax = transformed_xmax
    transformed_bbox_data.ymax = transformed_ymax
    transformed_bbox_data.keypoints = apply_perspective_transform_to_points(
        transformed_bbox_data.keypoints,
        perspective_matrix,
        result_height,
        result_height,
        allow_negative_and_large_coords,
        remove_bad_coords,
    )
    if isinstance(transformed_bbox_data.mask, list):
        transformed_bbox_data.mask = [
            apply_perspective_transform_to_points(
                polygon,
                perspective_matrix,
                result_height,
                result_height,
                allow_negative_and_large_coords,
                remove_bad_coords,
            )
            for polygon in transformed_bbox_data.mask
        ]
        transformed_bbox_data.mask = [polygon for polygon in transformed_bbox_data.mask if len(polygon) > 0]
    transformed_bbox_data.additional_bboxes_data = [
        _apply_perspective_transform_to_bbox_data(
            additional_bbox_data,
            perspective_matrix,
            result_height,
            result_height,
            allow_negative_and_large_coords,
            remove_bad_coords,
        )
        for additional_bbox_data in transformed_bbox_data.additional_bboxes_data
    ]
    transformed_bbox_data.additional_bboxes_data = [
        additional_bbox_data
        for additional_bbox_data in transformed_bbox_data.additional_bboxes_data
        if additional_bbox_data is not None
    ]
    transformed_bbox_data.cropped_image = None
    return transformed_bbox_data


def get_perspective_matrix_for_base_keypoints(
    base_keypoints: Tuple[Tuple[int, int], Tuple[int, int], Tuple[int, int], Tuple[int, int]]
) -> Tuple[np.ndarray, Tuple[int, int]]:
    base_keypoints = np.array(base_keypoints, dtype=np.float32)
    (top_left, top_right, bottom_right, bottom_left) = base_keypoints
    width_a = np.linalg.norm(bottom_right - bottom_left)
    width_b = np.linalg.norm(top_right - top_left)
    height_a = np.linalg.norm(top_right - bottom_right)
    height_b = np.linalg.norm(top_left - bottom_left)
    result_width = max(int(width_a), int(width_b))
    result_height = max(int(height_a), int(height_b))
    transformed_points = np.array(
        [[0, 0], [result_width - 1, 0], [result_width - 1, result_height - 1], [0, result_height - 1]], dtype=np.float32
    )
    perspective_matrix = cv2.getPerspectiveTransform(base_keypoints, transformed_points)
    return perspective_matrix, (result_width, result_height)


def apply_perspective_transform_to_image_data(
    image_data: ImageData,
    perspective_matrix: np.ndarray,
    result_width: int,
    result_height: int,
    allow_negative_and_large_coords: bool,
    remove_bad_coords: bool,
    open_image: bool = True,
    exif_transpose: bool = False,
) -> ImageData:
    image = image_data.open_image(returns_none_if_empty=True, exif_transpose=exif_transpose) if open_image else None
    image = cv2.warpPerspective(image, perspective_matrix, (result_width, result_height)) if image is not None else None

    image_data = copy.deepcopy(image_data)
    image_data.keypoints = apply_perspective_transform_to_points(
        image_data.keypoints,
        perspective_matrix,
        result_width,
        result_height,
        allow_negative_and_large_coords,
        remove_bad_coords,
    )
    if isinstance(image_data.mask, list):
        image_data.mask = [
            apply_perspective_transform_to_points(
                polygon,
                perspective_matrix,
                result_height,
                result_height,
                allow_negative_and_large_coords,
                remove_bad_coords,
            )
            for polygon in image_data.mask
        ]
        image_data.mask = [polygon for polygon in image_data.mask if len(polygon) > 0]
    transformed_bboxes_data = [
        _apply_perspective_transform_to_bbox_data(
            bbox_data,
            perspective_matrix,
            result_width,
            result_height,
            allow_negative_and_large_coords,
            remove_bad_coords,
        )
        for bbox_data in image_data.bboxes_data
    ]
    image_data.bboxes_data = [bbox_data for bbox_data in transformed_bboxes_data if bbox_data is not None]
    image_data.image_path = None
    image_data.image = image

    return image_data


def non_max_suppression_image_data(
    image_data: ImageData, iou: float, score_threshold: float = float("-inf")
) -> ImageData:
    """
    Taken from https://towardsdatascience.com/non-maxima-suppression-139f7e00f0b5
    """
    image_data = copy.deepcopy(image_data)
    boxes = np.array([bbox_data.coords for bbox_data in image_data.bboxes_data])
    if len(boxes) == 0:
        return image_data
    x1 = boxes[:, 0]  # x coordinate of the top-left corner
    y1 = boxes[:, 1]  # y coordinate of the top-left corner
    x2 = boxes[:, 2]  # x coordinate of the bottom-right corner
    y2 = boxes[:, 3]  # y coordinate of the bottom-right corner
    # compute the area of the bounding boxes and sort the bounding
    # boxes by the bottom-right y-coordinate of the bounding box
    areas = (x2 - x1 + 1) * (y2 - y1 + 1)  # We have a least a box of one pixel, therefore the +1
    indices = np.arange(len(boxes))
    for i, bbox_data in enumerate(image_data.bboxes_data):
        temp_indices = indices[indices != i]
        xx1 = np.maximum(bbox_data.xmin, boxes[temp_indices, 0])
        yy1 = np.maximum(bbox_data.ymin, boxes[temp_indices, 1])
        xx2 = np.minimum(bbox_data.xmax, boxes[temp_indices, 2])
        yy2 = np.minimum(bbox_data.ymax, boxes[temp_indices, 3])
        w = np.maximum(0, xx2 - xx1 + 1)
        h = np.maximum(0, yy2 - yy1 + 1)
        # compute the ratio of overlap
        overlap = (w * h) / areas[temp_indices]
        if np.any(overlap) > iou:
            indices = indices[indices != i]

    image_data.bboxes_data = [image_data.bboxes_data[i] for i in indices]
    if score_threshold is not None:
        image_data.bboxes_data = [
            bbox_data
            for bbox_data in image_data.bboxes_data
            if (bbox_data.detection_score is None) or (bbox_data.detection_score >= score_threshold)
        ]
    image_data.image_path = image_data.image_path
    image_data.image = image_data.image
    return image_data


def non_max_suppression_image_data_using_tf(
    image_data: ImageData, iou: float, score_threshold: float = float("-inf"), ignore_classes: bool = True
) -> ImageData:
    import tensorflow as tf

    image_data = copy.deepcopy(image_data)
    if len(image_data.bboxes_data) <= 1:
        return image_data
    if not ignore_classes:
        labels = sorted(set([bbox_data.label for bbox_data in image_data.bboxes_data]))
        indexes = []
        for label in labels:
            indexes_by_label = [i for i, bbox_data in enumerate(image_data.bboxes_data) if bbox_data.label == label]
            bboxes_data_by_label = [bbox_data for bbox_data in image_data.bboxes_data if bbox_data.label == label]
            bboxes_by_label = [
                (bbox_data.ymin, bbox_data.xmin, bbox_data.ymax, bbox_data.xmax) for bbox_data in bboxes_data_by_label
            ]
            scores_by_label = [
                bbox_data.detection_score if bbox_data.detection_score is not None else 1.0
                for bbox_data in bboxes_data_by_label
            ]
            result_by_label = tf.image.non_max_suppression(
                bboxes_by_label,
                scores_by_label,
                len(bboxes_data_by_label),
                iou_threshold=float(iou),
                score_threshold=float(score_threshold),
            )
            indexes.extend([indexes_by_label[i] for i in result_by_label.numpy()])
        image_data.bboxes_data = [image_data.bboxes_data[i] for i in sorted(indexes)]
    else:
        bboxes = [
            (bbox_data.ymin, bbox_data.xmin, bbox_data.ymax, bbox_data.xmax) for bbox_data in image_data.bboxes_data
        ]
        scores = [
            bbox_data.detection_score if bbox_data.detection_score is not None else 1.0
            for bbox_data in image_data.bboxes_data
        ]
        result = tf.image.non_max_suppression(
            bboxes,
            scores,
            len(image_data.bboxes_data),
            iou_threshold=float(iou),
            score_threshold=float(score_threshold),
        )
        image_data.bboxes_data = [image_data.bboxes_data[i] for i in result.numpy()]
    return image_data


def split_by_grid(
    size: Tuple[int, int],  # (width, height)
    n_rows: int,
    n_cols: int,
    x_window_size: int,
    y_window_size: int,
    x_offset: int,
    y_offset: int,
    minimum_size: float = 0.5,
) -> List[BboxData]:
    width, height = size
    bboxes_data = [
        BboxData(
            xmin=x_offset + i * x_window_size,
            ymin=y_offset + j * y_window_size,
            xmax=min(width, x_offset + (i + 1) * x_window_size),
            ymax=min(height, y_offset + (j + 1) * y_window_size),
        )
        for i in range(n_rows)
        for j in range(n_cols)
    ]
    bboxes_data = [
        bbox_data
        for bbox_data in bboxes_data
        if (
            (bbox_data.xmax - bbox_data.xmin >= minimum_size * x_window_size)
            and (bbox_data.ymax - bbox_data.ymin >= minimum_size * y_window_size)
        )
    ]
    return bboxes_data


def split_image_by_grid(
    image: np.ndarray,
    n_rows: int,
    n_cols: int,
    x_window_size: int,
    y_window_size: int,
    x_offset: int,
    y_offset: int,
    minimum_size: float = 0.5,
) -> List[BboxData]:
    height, width = image.shape[0:2]
    bboxes_data = split_by_grid(
        size=(width, height),  # (width, height)
        n_rows=n_rows,
        n_cols=n_cols,
        x_window_size=x_window_size,
        y_window_size=y_window_size,
        x_offset=x_offset,
        y_offset=y_offset,
        minimum_size=minimum_size,
    )
    for bbox_data in bboxes_data:
        bbox_data.image = image
    return bboxes_data


def split_image_data_by_grid(
    image_data: ImageData,
    n_rows: int,
    n_cols: int,
    x_window_size: int,
    y_window_size: int,
    x_offset: int,
    y_offset: int,
    allow_large_coords: bool = False,
    split_by_grid_minimum_size: float = 0.5,
    minimum_crop_intersection_area: float = 0.2,
    minimum_relative_size_of_inner_bboxes: float = 0.5,
) -> ImageData:
    image_data = copy.deepcopy(image_data)
    width, height = image_data.get_image_size()
    crops_bboxes_data = split_by_grid(
        size=(width, height),
        n_rows=n_rows,
        n_cols=n_cols,
        x_window_size=x_window_size,
        y_window_size=y_window_size,
        x_offset=x_offset,
        y_offset=y_offset,
        minimum_size=split_by_grid_minimum_size,
    )

    def crop_intersection_area(crop_bbox_data: BboxData, bbox_data: BboxData) -> float:
        inter_xmin = np.max([crop_bbox_data.xmin, bbox_data.xmin])
        inter_ymin = np.max([crop_bbox_data.ymin, bbox_data.ymin])
        inter_xmax = np.min([crop_bbox_data.xmax, bbox_data.xmax])
        inter_ymax = np.min([crop_bbox_data.ymax, bbox_data.ymax])
        inter_width = max(0, inter_xmax - inter_xmin)
        inter_height = max(0, inter_ymax - inter_ymin)
        intersection_area = inter_width * inter_height
        bbox_area = (bbox_data.xmax - bbox_data.xmin) * (bbox_data.ymax - bbox_data.ymin)
        if bbox_area == 0:
            return 0.0
        return intersection_area / bbox_area

    def apply_fix_bbox(crop_bbox_data: BboxData, bbox_data: BboxData):
        new_xmin = max(0, max(bbox_data.xmin, crop_bbox_data.xmin))
        new_ymin = max(0, max(bbox_data.ymin, crop_bbox_data.ymin))
        new_xmax = max(0, min(bbox_data.xmax, crop_bbox_data.xmax))
        new_ymax = max(0, min(bbox_data.ymax, crop_bbox_data.ymax))
        if (bbox_data.xmax - bbox_data.xmin == 0.0) or (bbox_data.ymax - bbox_data.ymin == 0.0):
            return None
        relative_size = min(
            (new_xmax - new_xmin) / (bbox_data.xmax - bbox_data.xmin),
            (new_ymax - new_ymin) / (bbox_data.ymax - bbox_data.ymin),
        )
        if relative_size < minimum_relative_size_of_inner_bboxes:
            return None
        bbox_data.xmin = new_xmin
        bbox_data.xmax = new_xmax
        bbox_data.ymin = new_ymin
        bbox_data.ymax = new_ymax
        bbox_data.keypoints = bbox_data.keypoints[
            (
                (bbox_data.keypoints[:, 0] >= crop_bbox_data.xmin)
                & (bbox_data.keypoints[:, 1] >= crop_bbox_data.ymin)
                & (bbox_data.keypoints[:, 0] <= crop_bbox_data.ymax)
                & (bbox_data.keypoints[:, 1] <= crop_bbox_data.ymax)
            )
        ]
        additional_bboxes_data = [
            apply_fix_bbox(crop_bbox_data, additional_bbox_data)
            for additional_bbox_data in bbox_data.additional_bboxes_data
            if crop_intersection_area(crop_bbox_data, additional_bbox_data) >= minimum_crop_intersection_area
        ]
        bbox_data.additional_bboxes_data = [
            additional_bbox_data for additional_bbox_data in additional_bboxes_data if additional_bbox_data is not None
        ]
        return bbox_data

    for crop_bbox_data in crops_bboxes_data:
        if not allow_large_coords:
            additional_bboxes_data = [
                apply_fix_bbox(crop_bbox_data, bbox_data)
                for bbox_data in copy.deepcopy(image_data.bboxes_data)
                if crop_intersection_area(crop_bbox_data, bbox_data) >= minimum_crop_intersection_area
            ]
            additional_bboxes_data = [bbox_data for bbox_data in additional_bboxes_data if bbox_data is not None]
            keypoints = copy.deepcopy(image_data.keypoints)[
                (
                    (image_data.keypoints[:, 0] >= crop_bbox_data.xmin)
                    & (image_data.keypoints[:, 1] >= crop_bbox_data.ymin)
                    & (image_data.keypoints[:, 0] <= crop_bbox_data.ymax)
                    & (image_data.keypoints[:, 1] <= crop_bbox_data.ymax)
                )
            ]
        else:
            additional_bboxes_data = copy.deepcopy(image_data.bboxes_data)
            keypoints = copy.deepcopy(image_data.keypoints)

        crop_bbox_data.additional_bboxes_data = additional_bboxes_data
        crop_bbox_data.keypoints = keypoints

    image_data.bboxes_data = crops_bboxes_data
    image_data.image_path = image_data.image_path  # apply to bboxes_data
    image_data.image = image_data.image  # apply to bboxes_data
    return image_data


def uncrop_bboxes_data(
    bboxes_data: List[BboxData],
    src_xmin: int,
    src_ymin: int,
    src_image: Optional[np.ndarray] = None,
    src_image_path: Optional[np.ndarray] = None,
    src_image_height: Optional[int] = None,
    src_image_width: Optional[int] = None,
) -> BboxData:
    bboxes_data = copy.deepcopy(bboxes_data)

    def _append_cropped_bbox_data_to_image_data(bbox_data: BboxData):
        bbox_data.keypoints[:, 0] += src_xmin
        bbox_data.keypoints[:, 1] += src_ymin
        bbox_data.xmin += src_xmin
        bbox_data.ymin += src_ymin
        bbox_data.xmax += src_xmin
        bbox_data.ymax += src_ymin
        bbox_data.image = src_image
        bbox_data.image_path = src_image_path
        bbox_data.cropped_image = None
        bbox_data.meta_height = src_image_height
        bbox_data.meta_width = src_image_width
        for additional_bbox_data in bbox_data.additional_bboxes_data:
            _append_cropped_bbox_data_to_image_data(additional_bbox_data)

    for bbox_data in bboxes_data:
        _append_cropped_bbox_data_to_image_data(bbox_data)

    return bboxes_data


def concat_images_data(
    image_data_a: ImageData,
    image_data_b: ImageData,
    background_color_a: Tuple[int, int, int, int] = None,
    background_color_b: Tuple[int, int, int, int] = None,
    thumbnail_size_a: Tuple[int, int] = None,
    thumbnail_size_b: Tuple[int, int] = None,
    how: Literal["horizontally", "vertically"] = "horizontally",
    mode: Literal["L", "RGB", "RGBA"] = "RGBA",
    background_edge_width: int = 3,
    between_edge_width: int = 0,
    exif_transpose: bool = False,
) -> ImageData:
    image_data_a = copy.deepcopy(image_data_a)
    image_data_b = copy.deepcopy(image_data_b)

    if image_data_a is None and image_data_b is not None:
        return image_data_b
    if image_data_a is not None and image_data_b is None:
        return image_data_a

    image_a = image_data_a.open_image(exif_transpose=exif_transpose)
    image_b = image_data_b.open_image(exif_transpose=exif_transpose)

    ha, wa = image_a.shape[:2]
    hb, wb = image_b.shape[:2]

    image = concat_images(
        image_a=image_a,
        image_b=image_b,
        background_color_a=background_color_a,
        background_color_b=background_color_b,
        thumbnail_size_a=thumbnail_size_a,
        thumbnail_size_b=thumbnail_size_b,
        how=how,
        mode=mode,
        background_edge_width=background_edge_width,
        between_edge_width=between_edge_width,
    )
    image_data_a_new_xmin, image_data_a_new_ymin = None, None
    image_data_b_new_xmin, image_data_b_new_ymin = None, None

    if how == "horizontally":
        max_height = np.max([ha, hb])
        min_ha = max_height // 2 - ha // 2
        max_ha = max_height // 2 + ha // 2
        min_hb = max_height // 2 - hb // 2
        max_hb = max_height // 2 + hb // 2
        image_data_a_new_xmin = 0
        image_data_a_new_ymin = min_ha
        image_data_a_new_xmax = wa
        image_data_a_new_ymax = max_ha
        image_data_b_new_xmin = wa + between_edge_width
        image_data_b_new_ymin = min_hb
        image_data_b_new_xmax = wa + between_edge_width + wb
        image_data_b_new_ymax = max_hb

    elif how == "vertically":
        max_width = np.max([wa, wb])
        min_wa = max_width // 2 - wa // 2
        max_wa = max_width // 2 + wa // 2
        min_wb = max_width // 2 - wb // 2
        max_wb = max_width // 2 + wb // 2
        image_data_a_new_xmin = min_wa
        image_data_a_new_ymin = 0
        image_data_a_new_xmax = max_wa
        image_data_a_new_ymax = ha
        image_data_b_new_xmin = min_wb
        image_data_b_new_ymin = ha + between_edge_width
        image_data_b_new_xmax = max_wb
        image_data_b_new_ymax = ha + between_edge_width + hb

    keypoints_a = image_data_a.keypoints
    keypoints_b = image_data_a.keypoints
    keypoints_a[:, 0] += image_data_a_new_xmin
    keypoints_a[:, 1] += image_data_a_new_ymin
    keypoints_b[:, 0] += image_data_b_new_xmin
    keypoints_b[:, 1] += image_data_b_new_ymin

    def _get_new_coords_for_bbox_data(bbox_data: BboxData, xmin: int, ymin: int):
        bbox_data.keypoints[:, 0] += xmin
        bbox_data.keypoints[:, 1] += ymin
        bbox_data.xmin += xmin
        bbox_data.ymin += ymin
        bbox_data.xmax += xmin
        bbox_data.ymax += ymin
        bbox_data.image = None
        bbox_data.image_path = None
        bbox_data.cropped_image = None
        for additional_bbox_data in bbox_data.additional_bboxes_data:
            _get_new_coords_for_bbox_data(additional_bbox_data, xmin, ymin)

    for bbox_data in image_data_a.bboxes_data:
        _get_new_coords_for_bbox_data(bbox_data, image_data_a_new_xmin, image_data_a_new_ymin)

    if "concat_images_data__image_data" not in [bbox_data.label for bbox_data in image_data_a.bboxes_data]:
        bbox_data_a_into = [
            BboxData(
                xmin=image_data_a_new_xmin,
                ymin=image_data_a_new_ymin,
                xmax=image_data_a_new_xmax,
                ymax=image_data_a_new_ymax,
                label="concat_images_data__image_data",
                additional_bboxes_data=[
                    bbox_data
                    for bbox_data in image_data_a.bboxes_data
                    if "concat_images_data__image_data" != bbox_data.label
                ],
            )
        ]
    else:
        bbox_data_a_into = []
    image_data_a.bboxes_data = [
        bbox_data for bbox_data in image_data_a.bboxes_data if "concat_images_data__image_data" == bbox_data.label
    ] + bbox_data_a_into

    for bbox_data in image_data_b.bboxes_data:
        _get_new_coords_for_bbox_data(bbox_data, image_data_b_new_xmin, image_data_b_new_ymin)
    if "concat_images_data__image_data" not in [bbox_data.label for bbox_data in image_data_b.bboxes_data]:
        bbox_data_b_into = [
            BboxData(
                xmin=image_data_b_new_xmin,
                ymin=image_data_b_new_ymin,
                xmax=image_data_b_new_xmax,
                ymax=image_data_b_new_ymax,
                label="concat_images_data__image_data",
                additional_bboxes_data=[
                    bbox_data
                    for bbox_data in image_data_b.bboxes_data
                    if "concat_images_data__image_data" != bbox_data.label
                ],
            )
        ]
    else:
        bbox_data_b_into = []
    image_data_b.bboxes_data = [
        bbox_data for bbox_data in image_data_b.bboxes_data if "concat_images_data__image_data" == bbox_data.label
    ] + bbox_data_b_into

    image_data = ImageData(
        image_path=None,
        image=image,
        bboxes_data=image_data_a.bboxes_data + image_data_b.bboxes_data,
        label=None,
        keypoints=np.concatenate([keypoints_a, keypoints_b], axis=0),
        additional_info={**image_data_a.additional_info, **image_data_b.additional_info},
    )

    return image_data


def get_all_bboxes_data_in_image_data(
    image_data: ImageData,
    additional_bboxes_data_depth: Optional[int] = None,
) -> List[BboxData]:
    bboxes_data = []

    def _append_bbox_data(bbox_data: BboxData, depth: int):
        if additional_bboxes_data_depth is not None and depth > additional_bboxes_data_depth:
            return
        bboxes_data.append(bbox_data)
        for additional_bbox_data in bbox_data.additional_bboxes_data:
            _append_bbox_data(additional_bbox_data, depth + 1)

    for bbox_data in image_data.bboxes_data:
        _append_bbox_data(bbox_data, depth=0)

    return bboxes_data


def flatten_additional_bboxes_data_in_image_data(
    image_data: ImageData,
    additional_bboxes_data_depth: Optional[int] = None,
    set_additional_bboxes_data_empty: bool = True,
) -> ImageData:
    image_data = copy.deepcopy(image_data)
    bboxes_data = []

    def _append_bbox_data(bbox_data: BboxData, depth: int):
        if additional_bboxes_data_depth is not None and depth > additional_bboxes_data_depth:
            return
        bboxes_data.append(bbox_data)
        for additional_bbox_data in bbox_data.additional_bboxes_data:
            _append_bbox_data(additional_bbox_data, depth + 1)
        if set_additional_bboxes_data_empty:
            bbox_data.additional_bboxes_data = []

    for bbox_data in image_data.bboxes_data:
        _append_bbox_data(bbox_data, depth=0)

    image_data.bboxes_data = bboxes_data
    return image_data


def find_closest_pair(arr1: np.ndarray, arr2: np.ndarray) -> np.ndarray:
    distances = np.sum((arr1[:, np.newaxis, :] - arr2[np.newaxis, :, :]) ** 2, axis=-1)
    return np.unravel_index(np.argmin(distances, axis=None), distances.shape)


def combine_polygons(polygons: List[np.ndarray]) -> np.ndarray:
    combined = []
    polygons = [np.array(polygon).reshape(-1, 2) for polygon in polygons]
    index_pairs = [[] for _ in range(len(polygons))]

    for i in range(1, len(polygons)):
        idx1, idx2 = find_closest_pair(polygons[i - 1], polygons[i])
        index_pairs[i - 1].append(idx1)
        index_pairs[i].append(idx2)

    for pass_num in range(2):
        if pass_num == 0:
            for i, indices in enumerate(index_pairs):
                if len(indices) == 2 and indices[0] > indices[1]:
                    indices = indices[::-1]
                    polygons[i] = polygons[i][::-1, :]

                polygons[i] = np.roll(polygons[i], -indices[0], axis=0)
                polygons[i] = np.concatenate([polygons[i], polygons[i][:1]])
                if i in [0, len(index_pairs) - 1]:
                    combined.append(polygons[i])
                else:
                    indices = [0, indices[1] - indices[0]]
                    combined.append(polygons[i][indices[0] : indices[1] + 1])
        else:
            for i in range(len(index_pairs) - 1, -1, -1):
                if i not in [0, len(index_pairs) - 1]:
                    indices = index_pairs[i]
                    adjusted_idx = abs(indices[1] - indices[0])
                    combined.append(polygons[i][adjusted_idx:])
    return combined


def combine_mask_polygons_to_one_polygon(mask: List[List[np.ndarray]]) -> np.ndarray:
    if len(mask) > 1:
        keypoints = np.concatenate(combine_polygons(mask), axis=0).reshape(-1, 2)
    elif len(mask) == 1:
        keypoints = mask[0]
    else:
        keypoints = []
    keypoints = np.array(keypoints).reshape(-1, 2)
    return keypoints



from typing import List

import numpy as np
from sklearn.metrics.pairwise import cosine_similarity


def get_minmax_order(embeddings: np.ndarray) -> List:
    similarity_matrix = cosine_similarity(embeddings)

    items = np.unravel_index(np.argmin(similarity_matrix, axis=None), shape=similarity_matrix.shape)
    items = list(items)

    while len(items) < len(similarity_matrix):
        new_index = np.argmin(np.max(similarity_matrix[items], axis=0))
        items.append(new_index)

    return items



import io
from typing import Tuple

import fsspec
import numpy as np
from PIL import Image

from cv_pipeliner.utils.images import EXIF_ORIENTATION_TO_METHOD


def _exif_transpose_and_get_image_size(image: Image.Image) -> Tuple[int, int]:
    exif = image.getexif()
    orientation = exif.get(0x0112)
    method = EXIF_ORIENTATION_TO_METHOD.get(orientation)
    if method is not None:
        old_width, old_height = image.size
        image_zeros = Image.fromarray(np.zeros((old_height, old_width))).transpose(method)
        return image_zeros.size
    return image.size


def get_image_size(filepath: "bytes_or_path", exif_transpose: bool = True) -> Tuple[int, int]:
    if isinstance(filepath, bytes):  # file-like object
        fhandle = io.BytesIO(filepath)
    elif isinstance(filepath, io.BytesIO):  # file-like object
        fhandle = filepath
    elif isinstance(filepath, fsspec.core.OpenFile):  # file-like object
        fhandle = filepath.__enter__()
    else:
        fhandle = fsspec.open(filepath, "rb").__enter__()

    image = Image.open(fhandle)  # lazy loading, so it's ok to open
    size = image.size
    if exif_transpose:
        size = _exif_transpose_and_get_image_size(image)
    return size
